from BANG.configuration import Config_reader
import matplotlib.pyplot as plt

'''
Example for running the code: Parameter estimation and diagnostic plots

In the test directory there should be a file "mock_data.npy". The following code
will try to fit that data. I do not recommend running it on CPU.

For any user the steps for using BANG are the following:

- Get your data, go to the file configuration.py and change the function read_user in 
  ordet to read your data.
  Data should be a N X 13 matrix with the following columns:
        - column 0  should be x in arcsecond
        - column 1  should be y in arcsecond
        - column 2  should be x in kpc
        - column 3  should be y in kpc
        - column 4  should be the logarithm of the surface brightness in M_sun/kpc^2
        - column 5  should be the line of sight velocity in km/s
        - column 6  should be the line of sight velocity dispersion in km/s
        - column 7  should be the M/L ratio in M_sun/L_sun
        - column 8  should be the error of the logarithm of the surface brightness in M_sun/kpc^2
        - column 9  should be the error of the line of sight velocity in km/s
        - column 10 should be the error of the line of sight velocity dispersion in km/s
        - column 11 should be the error of the M/L ratio in M_sun/L_sun
        - column 12 should contain 0.0 in the pixels that you do not want to use for any of 
                    the observed quantities, put 1.0 if you want to use the pixel.

    Up to now support only galaxies with M X M pixels (number of pixels in the x direction
    equal to the number of pixels in the y direction, upgrade will be available soon). So 
    please if you galaxy is of dimension M X L add columns or rows in order to make it M X M
    or L X L. The you can simply put 0.0 in column 12 such that the new mock pixels will not be 
    used in the parameter estimation.

    Consider also that all the data in you data file must be 1D array. So if you have a M X M 
    matrix remember to flatten it. 

    Suppose that your data have a total number of pixels N then if we define n = int(N**0.5):
        y (or x) should be an array like [-1,0,1,-1,0,1,-1,0,1] where the [-1,0,1] slice has
            dimension n and it is repeated n times
        x (or y) should be an array like [3,3,3, 4,4,4, 5,5,5] where you have n different points
            each one repeated n times. 
    
    Note that it is not necessary to have your grid centered in zero.

    Use np.nan in order to mask the data that you do not want to use. For example if you do
    not have M/L ratio measurements simply put in column 7 all np.nan. Or put use_ML = False
    in your config file.

- Look at the config.yaml file and change all the parameters according the model that you want 
  to run. Check the components that you want to include, the ranges for the parameter search 
  and the input/output directories. All parameters are explained in the config.yaml file.

- run a script similar to the one below and wait for the result. On a standard gpu you parameter
  estimation should last ~ 1h. 

Please if you want to try band on the mock_data.npy file go to the test directory and launch the 
command:
                python3 example_script.py

All the useful function should be explained in their docscring.

For any problem in running BANG or in creating the data structure feel free to write to:
    
                            frigamonti@uninsubria.it

'''

if __name__=='__main__':

    # Configuration file name
    config_name = 'config.yaml'

    # Read configuration
    config = Config_reader(config_name)
    
    # Start parameter estimation
    config.parameter_estimation()

    # Plot corner plot with all the posterior probabilities
    corner = config.diagnostic(show=False)

    # Plot all the parameters tried during the fit
    iteration = config.diagnostic_iteration(show=False)
    
    # Compute 2D model with the best fit parameters
    B_model,v_model,sigma_model,ML_model,maps_2D = config.maps_2D(show=False,lim='model')

    # Compute 1D model with the best fit parameters
    maps_1D = config.maps_1D_all_outputs(show=False)
    
    # Show all plots
    corner.show()
    maps_2D.show()
    maps_1D.show()
    iteration.show()
    plt.show()

    # Uncomment following section for plotting model with specific set of parameters
    #Mb,Rb,Md1,Rd1,Md2,Rd2,logf,c = 10.55,0.3,10.1,0.4,10.5,1,1.716,12.966
    #x0,y0,theta,incl = -0.122,-0.063,2.691,1.109
    #k1,k2,ML_b,ML_d1,ML_d2 = 0.861,0.894,4.0,2.,1.
    #gamma = 1.7
    #sys_rho = -1.0
    #params = {'log10_bulge_mass':Mb,
    #           'log10_disc1_mass':Md1,
    #           'log10_disc2_mass':Md2,
    #           'log10_halo_fraction':logf,
    #           'log10_bulge_radius':Rb,
    #           'log10_disc1_radius':Rd1,
    #           'log10_disc2_radius':Rd2,
    #           'concentration':c,
    #           'x0':x0,
    #           'y0':y0,
    #           'orientation':theta,
    #           'inclination':incl,
    #           'ML_b':ML_b,
    #           'ML_d1':ML_d1,
    #           'ML_d2':ML_d2,
    #           'k1':k1,
    #           'k2':k2,
    #           'gamma':gamma,
    #           'sys_rho':sys_rho}

    #B_model,v_model,sigma_model,ML_model,maps_2D = config.maps_2D_specify_params(params,savename='2D_maps_mine.png',show=False,lim='model')

    #maps_1D = config.maps_1D_specify_params(params,show=False,savename='1D_maps_mine.png')

    #corner.show()
    #maps_2D.show()
    #maps_1D.show()
    #iteration.show()
    #plt.show()



import yaml
from BANG.parameter_search import GalaxyModel
import BANG.data as d
import numpy as np 
import cpnest
from BANG.data import data
from BANG.result_analysis import Result_Analysis
from BANG.model_creation import model 
import matplotlib.pyplot as plt
from astropy.cosmology import FlatLambdaCDM
import os

def read_yaml(file_path):
    ''' 
    Function
    ----------------------------------------------------------------------
    ----------------------------------------------------------------------
    read yaml (i.e. configuration) file
        
        Parameters:
            file_path: str
                absolute path to read
        Returns:
    '''
    with open(file_path, "r") as f:
        return yaml.safe_load(f)

def read_user(dataset_path,use_ML):
    ''' 
    Function
    ----------------------------------------------------------------------
    ----------------------------------------------------------------------
    read data (i.e. configuration) file. The user should adapt it to its 
    own dataset
        
        Parameters:
            dataset_path: str
                absolute path of the data file
            use_ML: bool
                If False then ML *= np.nan such that it is not 
                used in the analysis.
        Returns:
            x_arcsec: 1D arr float32
                x positions in arcsecond
            y_arcsec: 1D arr float32
                y positions in arcsecond
            x_kpc: 1D arr float32
                x positions in kpc
            y_kpc: 1D arr float32
                y positions in kpc
            ibrightness_data: 1D arr float32
                log10 brightness (i-band in this case).
                Use np.nan for masking pixels.
            v_data: 1D arr float32
                line of sight velocity data.
                Use np.nan for masking pixels.
            sigma_data: 1D arr float32
                line of sight velocity dispersion data.
                Use np.nan for masking pixels.
            ML_data: 1D arr float32
                Mass to light ratio data.
                Use np.nan for masking pixels.
            ibrightness_error: 1D arr float32
                log10 brightness errror (i-band in this case).
                Use np.nan for masking pixels.
            v_error: 1D arr float32
                line of sight velocity data errror.
                Use np.nan for masking pixels.
            sigma_error: 1D arr float32
                line of sight velocity dispersion data errror.
                Use np.nan for masking pixels.
            ML_error: 1D arr float32
                Mass to light ratio data errror.
                Use np.nan for masking pixels.
            goodpix: 1D arr float32
                If goodpix[i] == 0. then pixel "i" is not considered.
    '''
    a = np.load(dataset_path)
    x_arcsec = a[:,0]
    y_arcsec = a[:,1]
    x_kpc = a[:,2]
    y_kpc = a[:,3]
    ibrightness_data  = a[:,4]
    v_data            = a[:,5]
    sigma_data        = a[:,6]
    ML_data           = a[:,7]
    ibrightness_error = a[:,8]
    v_error           = a[:,9]
    sigma_error       = a[:,10]
    ML_error          = a[:,11]
    goodpix           = a[:,12]   
    SN                = a[:,13]

    if use_ML:
        pass
    else:
        ML_data *= np.nan
        ML_error *= np.nan
    
    return x_arcsec,y_arcsec,x_kpc,y_kpc,ibrightness_data,v_data,sigma_data,ML_data,ibrightness_error,v_error,sigma_error,ML_error,goodpix 

class Config_reader(data,Result_Analysis,model):
    def __init__(self,config_path,gpu_device=0):
        ''' 
        A class for reading configuration files and performing parameter estimation and all diagnostic plots.
        ----------------------------------------------------------------------
        ----------------------------------------------------------------------                                           
            Attributes:
                    config_path: str
                        name/path of the configuration file
                    gpu_device: int16
                        integer referring to the gpu machine. Useful only
                        in case of multi-gpu usage
            Methods:
                parameter_estimation(gpu_device=0):
                    Compute parameter estimation.
                diagnostic(savename='corner.png',show=True,drop_sys_rho=False,drop_halo=False)
                    Plot corner plot of the whole parameter space
                diagnostic_iteration(savename='iteration.png',show=True,drop_sys_rho=False,drop_halo=False)
                    Plot each parameter as a function of iterations
                maps_2D(savename='2D_maps.png',plot_ML=True,show=True,percentile=[50,16,84],write=True,lim='data',savedata='2D_informations.npy',vmin=None,vmax=None,close=False)
                    Plot best fit model (2D maps)
                maps_1D(savename='1D_profiles.png',plot_ML=True,show=True,percentile=[50,16,84],write=True,close=False)
                    Plot best fit model (1D maps, averaged profiles)
                maps_1D_all_outputs(savename='1D_profiles.png',plot_ML=True,show=True,percentile=[50,16,84],savedata='1D_informations.npy',write=True,close=False)
                    Plot best fit model with all the components(1D maps, averaged profiles)
                maps_2D_specify_params(Best_fit_params={},savename='2D_maps.png',plot_ML=True,show=True,write=True,lim='data',vmin=None,vmax=None,close=False)
                    Plot model with using Best_fit_params as parameters (2D maps)
                maps_1D_specify_params(Best_fit_params={},savename='1D_profiles.png',plot_ML=True,show=True,write=True,close=False)
                    Plot model with using Best_fit_params as parameters (1D maps, averaged profiles)
                best_params_all_fits(write=True,filename='all_params.txt')
                    Write best fit params to file while fitting multiples objects
                mass_plot(min_r=0.1,max_r=10,Nbins=20,figname='mass_profile.png')  
                    Plot mass profile, actually available only with Bulge + inner disc+ outer disc model  
        '''
        #super().__init__(config_path)
        #super(Config_reader,self).__init__()
        config = read_yaml(config_path)

        if "SNR_threshold_velocity" in config["Settings"]:
            self.SNR_threshold_velocity = config["Settings"]["SNR_threshold_velocity"]
        if "SNR_threshold_dispersion" in config["Settings"]:
            self.SNR_threshold_dispersion = config["Settings"]["SNR_threshold_dispersion"]
        if "SNR_threshold_ML" in config["Settings"]:
            self.SNR_threshold_ML = config["Settings"]["SNR_threshold_ML"]
        else:
            self.SNR_threshold = 0.
        if "velocity_offset" in config["Settings"]:
            self.vel_offset = config["Settings"]["velocity_offset"]
        else:
            self.vel_offset = False
        if "use_ML" in config["Settings"]:
            self.use_ML = config["Settings"]["use_ML"]
        else:
            self.use_ML = True
            
        self.redshift = config["galaxy"]["redshift"]
        self.galaxy_name = config["galaxy"]["Name"]
        # READ device and fastmath option
        self.device = config["device"]["name"]
        self.fast_math = config["device"]["fastmath"]
        
        # READ input file directory and output directory
        self.input_path = config["Settings"]["input_path"]
        self.output_path = config["Settings"]["output_path"]
        self.output_path += '/'
        # READ psf sigma in kpc
        self.kin_sigma_psf = config["Settings"]['kin_psf_sigma_kpc']
        self.pho_sigma_psf = config["Settings"]['pho_psf_sigma_kpc']
        self.N_px = config["Settings"]['N_pixels']
        self.resolution = config["Settings"]['resolution']

        galactic_components = config["galactic_components"]
        reference_frame     = config["Reference_frame"]
        systematic_errors   = config["systematic_errors"]

        # download all parameter names 
        names,LaTex_names,bounds,value,vary,self.include  = [],[],[],[],[],[]
        self.bulge_type = galactic_components['Bulge']['name']
        self.halo_type = galactic_components['Halo']['name']

        for k in galactic_components:
            # Read which component to include
            self.include.append(galactic_components[k]['include'])
            if galactic_components[k]['include'] == True:
                # choose parameters
                for j in galactic_components[k]["parameters"]:
                    names.append(galactic_components[k]["parameters"][j]['name'])
                    LaTex_names.append(galactic_components[k]["parameters"][j]['LaTex_name'])
                    vary.append(galactic_components[k]["parameters"][j]['vary'])
                    bounds.append(galactic_components[k]["parameters"][j]['bounds'])
                    value.append(galactic_components[k]["parameters"][j]['value'])
            else:
                # exclude component
                pass

        for k in reference_frame:
            names.append(reference_frame[k]["name"])
            LaTex_names.append(reference_frame[k]["LaTex_name"])
            bounds.append(reference_frame[k]["bounds"])
            vary.append(reference_frame[k]["vary"])
            value.append(reference_frame[k]["value"])  

        for k in systematic_errors:
            names.append(systematic_errors[k]["name"])
            LaTex_names.append(systematic_errors[k]["LaTex_name"])
            bounds.append(systematic_errors[k]["bounds"])
            vary.append(systematic_errors[k]["vary"])
            value.append(systematic_errors[k]["value"])

        if config["Settings"]["sys_rho"] == True:
            pass 
        else:
            print("Error likelihood without systematic error is not implemented")
            print("Put fixed=True and value=-10 in systematic_errors")
            exit()


        # Build all the relevant dictionaries
        self.variable_quantities = {names[i]:bounds[i] for i in range(0,len(names)) if vary[i]==True}
        self.fixed_quantities = {names[i]:value[i] for i in range(0,len(names)) if vary[i]==False}
        self.variable_LaTex = {names[i]:LaTex_names[i] for i in range(0,len(names)) if vary[i]==True}
        self.fixed_LaTex = {names[i]:LaTex_names[i] for i in range(0,len(names)) if vary[i]==False}


        # READ dataset
        self.x_arcsec,self.y_arcsec,self.x,self.y,self.ibrightness_data,self.v_data,self.sigma_data,self.ML_data,\
        self.ibrightness_error,self.v_error,self.sigma_error,self.ML_error,self.goodpix = read_user(self.input_path,velocity_offset=self.vel_offset,\
                                                                                                    use_ML=self.use_ML,SNR_threshold_velocity=self.SNR_threshold_velocity,\
                                                                                                    SNR_threshold_dispersion=self.SNR_threshold_dispersion,\
                                                                                                    SNR_threshold_ML=self.SNR_threshold_ML)

        #####################################################################
        #####################################################################
        # CREATE data_object useful for later plots
        #####################################################################
        #####################################################################
        self.data_obj = data(self.x,self.y)
        # 0.009 assumes FWHM of 0.3 arcsec
        #self.x_grid,self.y_grid,self.psf = self.data_obj.refined_grid(size=int(np.sqrt(self.K)),sigma_psf=self.sigma_psf,sampling='linear',N_psf=4)
        self.x_grid,self.y_grid,self.kin_psf,self.K,self.indexing = self.data_obj.refined_grid_indexing(sigma_psf=self.kin_sigma_psf,N_px=self.N_px,resolution=self.resolution)
        _,_,self.pho_psf,_,_ = self.data_obj.refined_grid_indexing(sigma_psf=self.pho_sigma_psf,N_px=self.N_px,resolution=self.resolution)

        self.arcsec_to_kpc = abs(self.x[0]-self.x[1])/abs(self.x_arcsec[0]-self.x_arcsec[1])
        #####################################################################
        #####################################################################
        # CREATE model_object useful for later plots
        #####################################################################
        #####################################################################
        cosmo = FlatLambdaCDM(H0=70, Om0=0.3)
        H_z   = cosmo.H(self.redshift).value # this is in km/Mpc/s
        H_z   = H_z*1e5/(1e6*3.086e18)         # this is in 1/s
        G     = 4.299e-6                       # kcp*km^2/M_sun/s^2
        G_1   = G * (1e5*1e-3*3.24078e-19)**2  # kcp^3/M_sun/s^2
        self.rho_crit = 3*H_z*H_z/(8*np.pi*G_1)

        self.model_obj = model(self.device,
                              gpu_device,
                              self.fast_math,
                              self.include,
                              self.bulge_type,
                              self.halo_type,
                              self.data_obj.N,
                              self.data_obj.J,
                              self.K,
                              self.x_grid,
                              self.y_grid,
                              self.ibrightness_data,
                              self.v_data,
                              self.sigma_data,
                              self.ML_data,
                              self.ibrightness_error,
                              self.v_error,
                              self.sigma_error,
                              self.ML_error,
                              self.kin_psf,
                              self.pho_psf,
                              self.indexing,
                              self.goodpix,
                              self.rho_crit)
        #####################################################################
        #####################################################################
        # CREATE result_object useful for later plots
        #####################################################################
        #####################################################################
        self.result_obj = Result_Analysis(self.output_path,
                                          self.rho_crit,
                                          self.halo_type)

        #####################################################################
        #####################################################################
        # READ SPECIFIC OF NESTED SAMPLING
        #####################################################################
        #####################################################################
        self.verbose      = config["parameter_estimator"]["Nested_sampling"]["verbose"]
        self.poolsize     = config["parameter_estimator"]["Nested_sampling"]["poolsize"]
        self.nthreads     = config["parameter_estimator"]["Nested_sampling"]["nthreads"]
        self.nlive        = config["parameter_estimator"]["Nested_sampling"]["nlive"]
        self.maxmcmc      = config["parameter_estimator"]["Nested_sampling"]["maxmcmc"]
        self.nhamiltonian = config["parameter_estimator"]["Nested_sampling"]["nhamiltonian"]
        self.nslice       = config["parameter_estimator"]["Nested_sampling"]["nslice"]
        self.resume       = config["parameter_estimator"]["Nested_sampling"]["resume"]

    def parameter_estimation(self,gpu_device=0):
        ''' 
        Class Method
        ----------------------------------------------------------------------
        ----------------------------------------------------------------------
        Start parameter estimation assuming model specified in configuration file
            
            Parameters:
                gpu_device: int16
                    ID of the GPU. If multi-gpu is not needed, set it to zero.

            Returns:

        '''
        M = GalaxyModel(self.device,
                        gpu_device,
                        self.fast_math,
                        self.include,
                        self.bulge_type,
                        self.halo_type,
                        self.data_obj.N,
                        self.data_obj.J,
                        self.K,
                        self.x_grid,
                        self.y_grid,
                        self.ibrightness_data,
                        self.v_data,
                        self.sigma_data,
                        self.ML_data,
                        self.ibrightness_error,
                        self.v_error,
                        self.sigma_error,
                        self.ML_error,
                        self.kin_psf,
                        self.pho_psf,
                        self.indexing,
                        self.goodpix,
                        self.variable_quantities,  
                        self.fixed_quantities,
                        self.rho_crit)

        work=cpnest.CPNest(M,
                           verbose  = self.verbose,
                           poolsize = self.poolsize,
                           nthreads = self.nthreads ,
                           nlive    = self.nlive,
                           maxmcmc  = self.maxmcmc,
                           output   = self.output_path,
                           nhamiltonian = self.nhamiltonian,
                           nslice   = self.nslice,
                           resume   = self.resume,
                           periodic_checkpoint_interval = 600)
        work.run()

        work.get_posterior_samples()
        work.get_nested_samples()

    def diagnostic(self,savename='corner.png',show=True,drop_sys_rho=False,drop_halo=False):
        ''' 
        Class Method
        ----------------------------------------------------------------------
        ----------------------------------------------------------------------
        Plot posterior probabilities of all the parameters, assume existence of 
        "posterior.dat" file.
            
            Parameters:
                savename: str
                    Name for saving the plot
                show: Bool
                    If plot should be shown on screen
                drop_sys_rho: Bool
                    If used and set to True put systematic_error of the brightnees in 
                    cornern plot
                drop_halo: Bool
                    If True does not show halo posteriors

            Returns:
                fig: figure
                    corner plot
        '''
        fig = self.result_obj.corner_plot(self.variable_LaTex)
        fig.savefig(self.output_path+savename)
        if show==True:
            fig.show()
            plt.show()
        return fig
    
    def diagnostic_iteration(self,savename='iteration.png',show=True,drop_sys_rho=False,drop_halo=False):
        ''' 
        Class Method
        ----------------------------------------------------------------------
        ----------------------------------------------------------------------
        Plot parameters as a function of iteration in the nested sampling, assume existence of 
        "nested_samples.dat" file.
            
            Parameters:
                savename: str
                    Name for saving the plot
                show: Bool
                    If plot should be shown on screen
                drop_sys_rho: Bool
                    If used and set to True put systematic_error of the brightnees in 
                    cornern plot
                drop_halo: Bool
                    If True does not show halo posteriors

            Returns:
                fig: figure
                    corner plot
        '''
        fig = self.result_obj.iteration_plot(self.variable_LaTex)
        fig.savefig(self.output_path+savename)
        if show==True:
            fig.show()
            plt.show()
        return fig

    def maps_2D(self,savename='2D_maps.png',plot_ML=True,show=True,percentile=[50,16,84],write=True,lim='data',savedata='2D_informations.npy',vmin=None,vmax=None,close=False):
        ''' 
        Class Method
        ----------------------------------------------------------------------
        ----------------------------------------------------------------------
        Plot best fit model assuming existence of "posterior.dat" file. Plot is 
        3 X 4 or 3 X 3. 
        First line is the best fit model, second line are the data and third line are the residuals.
        From left to right: brightness,velocity,dispersion and (possibly) Mass to light ratio.
            
            Parameters:
                savename: str
                    Name for saving the plot
                plot_ML: str
                    If True plot also Mass to light ratio
                show: Bool
                    If plot should be shown on screen
                percentile: 1D arr float32 
                    Which "percentiles" to use for computing best fit params and error.
                    Default is:
                            - 50 (median) for best fit params
                            - 16 for lower error
                            - 84 for upper error
                write: Bool
                    Whether or not save to file the best fit params and errors.
                lim: str
                    Wheter to use "data" or "model" or "user" for the limits of the plot
                vmin: 1D arr float32
                    In case lim is set to "user" specify min value of the colormap
                vmax: 1D arr float32
                    In case lim is set to "user" specify max value of the colormap
                savedata: str
                    name of the file for saving best fit model values. Data are saved in the 
                    following format:
                        -tot[:,0] : x in arcsec
                        -tot[:,1] : y in arcsec
                        -tot[:,2] : x in kpc
                        -tot[:,3] : y in kpc
                        -tot[:,4] : brightness data (log10)
                        -tot[:,5] : brightness model (log10)
                        -tot[:,6] : v data
                        -tot[:,7] : v model
                        -tot[:,8] : sigma data
                        -tot[:,9] : sigma model
                        -tot[:,10] : ML data
                        -tot[:,11] : ML model
                        -tot[:,12] : brightness log10 error
                        -tot[:,13] : v data error
                        -tot[:,14] : sigma data error
                        -tot[:,15] : ML data error

            Returns:
                ibrightness_model: 1D arr float32
                    best fit brightness
                v_model: 1D arr float32
                    best fit velocity
                sigma_model: 1D arr float32
                    best fit dispersion
                ML_model: 1D arr float32
                    best fit mass to ligh
                fig: figure

        '''
        Best_fit_params,_,_ = self.result_obj.best_params(percentile=percentile,write=write)
        all_params = {**Best_fit_params, **self.fixed_quantities}
        ibrightness_model,_,_,_,v_model,_,_,_,sigma_model,_,_,_,LM_ratio,_,_,_ = self.model_obj.model(all_params)
        ML_model = 1.0/LM_ratio
        #ML_model = np.log10(1.0/LM_ratio)
        fig = self.data_obj.all_maps_kpc(self.ibrightness_data,
                                    np.log10(ibrightness_model),
                                    self.v_data,
                                    v_model,
                                    self.sigma_data,
                                    sigma_model,
                                    self.ML_data,
                                    ML_model,
                                    10**self.ibrightness_data*self.ibrightness_error*np.log(10),
                                    self.v_error,
                                    self.sigma_error,
                                    self.ML_error,
                                    plot_ML=plot_ML,
                                    lim = lim,
                                    vmin=vmin,
                                    vmax=vmax)
        fig.savefig(self.output_path+savename)
        if show==True:
            fig.show()
            plt.show()
        tot = np.concatenate((self.x_arcsec.reshape(-1,1),
                              self.y_arcsec.reshape(-1,1),
                              self.x.reshape(-1,1),
                              self.y.reshape(-1,1),
                              self.ibrightness_data.reshape(-1,1),
                              np.log10(ibrightness_model).reshape(-1,1),
                              self.v_data.reshape(-1,1),
                              v_model.reshape(-1,1),
                              self.sigma_data.reshape(-1,1),
                              sigma_model.reshape(-1,1),
                              self.ML_data.reshape(-1,1),
                              ML_model.reshape(-1,1),
                              self.v_error.reshape(-1,1),
                              self.sigma_error.reshape(-1,1),
                              self.ML_error.reshape(-1,1),
                              ),axis=1)
        np.save(self.output_path+savedata,tot)

        return ibrightness_model,v_model,sigma_model,ML_model,fig
        
    def maps_1D(self,savename='1D_profiles.png',plot_ML=True,show=True,percentile=[50,16,84],write=True,close=False):
        ''' 
        Class Method
        ----------------------------------------------------------------------
        ----------------------------------------------------------------------
        Plot best fit model assuming existence of "posterior.dat" file. Plot is 
        1 X 4 or 1 X 3. 
        From left to right: brightness,velocity,dispersion and (possibly) Mass to light ratio.
        Profiles are azimuthally averaged.

            Parameters:
                savename: str
                    Name for saving the plot
                plot_ML: str
                    If True plot also Mass to light ratio
                show: Bool
                    If plot should be shown on screen
                percentile: 1D arr float32 
                    Which "percentiles" to use for computing best fit params and error.
                    Default is:
                            - 50 (median) for best fit params
                            - 16 for lower error
                            - 84 for upper error
                write: Bool
                    Whether or not save to file the best fit params and errors.
            Returns:
                fig: figure

        '''
        Best_fit_params,_,_ = self.result_obj.best_params(percentile=percentile,write=write)
        all_params = {**Best_fit_params , **self.fixed_quantities}
        ibrightness_model,_,_,_,v_model,_,_,_,sigma_model,_,_,_,LM_ratio,_,_,_ = self.model_obj.model(all_params)

        ML_model = 1.0/LM_ratio
        #ML_model = np.log10(1.0/LM_ratio)
        minr = 0.4  # dimension of pixel is 0.5 arcsec
        maxr_kin = np.nanmax((self.x_arcsec**2+self.y_arcsec**2)**0.5*self.v_data/self.v_data)
        maxr_phot = 1.2*maxr_kin
        maxr_kin -= 1.5
        Nbin_phot = int((maxr_phot-minr)/1.0)
        Nbin_kin  = int((maxr_kin-minr)/1.0)

        Nbin_phot = max(Nbin_phot,7) 
        Nbin_kin  = max(Nbin_kin,7) 


        r_avg_B,B_avg_model  = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,ibrightness_model,quantity=None,estimator='mean',
                                   min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,B_avg_data         = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,10**self.ibrightness_data,quantity=None,estimator='mean',
                                   min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,B_avg_err          = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,10**self.ibrightness_data*self.ibrightness_error*np.log(10),quantity=None,estimator='mean',
                                   min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        
        r_avg_v,v_avg_model  = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,v_model,quantity="velocity",estimator='mean',
                                   min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=False)
        _,v_avg_data         = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.v_data,quantity="velocity",estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=False)
        _,v_avg_err          = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.v_error,quantity="velocity",estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=False)
        
        r_avg_sigma,sigma_avg_model = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,sigma_model,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=True)
        _,sigma_avg_data            = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.sigma_data,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=True)
        _,sigma_avg_err            = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.sigma_error,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=True)
        
        r_avg_ML,ML_avg_model = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,ML_model,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,ML_avg_data         = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.ML_data,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,ML_avg_err          = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.ML_error,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)

        #B_avg_err = np.nanmean(10**self.ibrightness_data*self.ibrightness_error*np.log(10))
        #v_avg_err = np.nanmean(self.v_error)
        #sigma_avg_err = np.nanmean(self.sigma_error)
        #ML_avg_err = np.nanmean(10**self.ML_data*self.ML_error*np.log(10))
        #ML_avg_err = np.nanmean(self.ML_error)

        fig = self.data_obj.all_1D_profile(
                                      self.arcsec_to_kpc,
                                      r_avg_B,
                                      B_avg_data,
                                      B_avg_model,
                                      r_avg_v,
                                      v_avg_data,
                                      v_avg_model,
                                      r_avg_sigma,
                                      sigma_avg_data,
                                      sigma_avg_model,
                                      r_avg_ML,
                                      ML_avg_data,
                                      ML_avg_model,
                                      B_avg_err,
                                      v_avg_err,
                                      sigma_avg_err,
                                      ML_avg_err,
                                      plot_ML=plot_ML)
        
        fig.savefig(self.output_path+savename)
        if show==True:
            fig.show()
            plt.show()

        return fig

    def maps_1D_all_outputs(self,savename='1D_profiles.png',plot_ML=True,show=True,percentile=[50,16,84],savedata='1D_informations.npy',write=True,close=False):
        '''
        Class Method
        ----------------------------------------------------------------------
        ----------------------------------------------------------------------
        Plot best fit model assuming existence of "posterior.dat" file. Plot is 
        1 X 4 or 1 X 3. 
        From left to right: brightness,velocity,dispersion and (possibly) Mass to light ratio.
        Profiles are azimuthally averaged. All components are present.
            
            Parameters:
                savename: str
                    Name for saving the plot
                plot_ML: str
                    If True plot also Mass to light ratio
                show: Bool
                    If plot should be shown on screen
                percentile: 1D arr float32 
                    Which "percentiles" to use for computing best fit params and error.
                    Default is:
                            - 50 (median) for best fit params
                            - 16 for lower error
                            - 84 for upper error
                write: Bool
                    Whether or not save to file the best fit params and errors.
                savedata: str
                    name of the file for saving best fit model values. Data are saved in the 
                    
                    see below the format
                       
            Returns:
                ibrightness_model: 1D arr float32
                    best fit brightness
                v_model: 1D arr float32
                    best fit velocity
                sigma_model: 1D arr float32
                    best fit dispersion
                ML_model: 1D arr float32
                    best fit mass to ligh
                fig: figure

        Note: this is probably working only with bulge+ inner disc+ outer disc model
        '''
        Best_fit_params,_,_ = self.result_obj.best_params(percentile=percentile,write=write)
        all_params = {**Best_fit_params , **self.fixed_quantities}
        ibrightness_model,B_B,B_D1,B_D2,v_model,v_B,v_D1,v_D2,\
        sigma_model,sigma_B,sigma_D1,sigma_D2,LM_ratio,LM_B,LM_D1,LM_D2 = self.model_obj.model(all_params)
        ML_model,ML_B,ML_D1,ML_D2 = 1.0/LM_ratio,1.0/LM_B,1.0/LM_D1,1.0/LM_D2
        #ML_model = np.log10(1.0/LM_ratio)

        minr = 0.4  # dimension of pixel is 0.5 arcsec
        maxr_kin = np.nanmax((self.x_arcsec**2+self.y_arcsec**2)**0.5*self.v_data/self.v_data)
        maxr_phot = 1.2**maxr_kin
        maxr_kin -= 1.5
        Nbin_phot = int((maxr_phot-minr)/1.0)
        Nbin_kin  = int((maxr_kin-minr)/1.0)

        Nbin_phot = max(Nbin_phot,7) 
        Nbin_kin  = max(Nbin_kin,7) 

        r_avg_B,B_avg_model  = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,np.log10(ibrightness_model),quantity=None,estimator='mean',
                                   min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,B_B_avg   = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,np.log10(B_B),quantity=None,estimator='mean',
                                   min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,B_D1_avg  = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,np.log10(B_D1),quantity=None,estimator='mean',
                                   min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,B_D2_avg  = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,np.log10(B_D2),quantity=None,estimator='mean',
                                   min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,B_avg_data         = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.ibrightness_data,quantity=None,estimator='mean',
                                   min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,B_avg_err          = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.ibrightness_error,quantity=None,estimator='mean',
                                   min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        
        r_avg_v,v_avg_model  = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,v_model,quantity="velocity",estimator='mean',
                                   min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=False)
        _,v_B_avg   = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,v_B,quantity="velocity",estimator='mean',
                                   min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=False)
        _,v_D1_avg  = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,v_D1,quantity="velocity",estimator='mean',
                                   min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=False)
        _,v_D2_avg  = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,v_D2,quantity="velocity",estimator='mean',
                                   min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=False)
        _,v_avg_data         = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.v_data,quantity="velocity",estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=False)
        _,v_avg_err          = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.v_error,quantity="velocity",estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=False)
        
        r_avg_sigma,sigma_avg_model = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,sigma_model,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=True)
        _,sigma_B_avg = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,sigma_B,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=True)
        _,sigma_D1_avg = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,sigma_D1,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=True)
        _,sigma_D2_avg = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,sigma_D2,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=True)
        _,sigma_avg_data            = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.sigma_data,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=True)
        _,sigma_avg_err            = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.sigma_error,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=True)
        
        r_avg_ML,ML_avg_model = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,ML_model,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,ML_B_avg = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,ML_B,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,ML_D1_avg = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,ML_D1,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,ML_D2_avg = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,ML_D2,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,ML_avg_data         = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.ML_data,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,ML_avg_err          = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.ML_error,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)

        #B_avg_err = np.nanmean(10**self.ibrightness_data*self.ibrightness_error*np.log(10))
        #v_avg_err = np.nanmean(self.v_error)
        #sigma_avg_err = np.nanmean(self.sigma_error)
        #ML_avg_err = np.nanmean(10**self.ML_data*self.ML_error*np.log(10))
        #ML_avg_err = np.nanmean(self.ML_error)

        fig = self.data_obj.all_1D_profile_all_outputs(
                                      self.arcsec_to_kpc,
                                      r_avg_B,
                                      B_avg_data,
                                      B_avg_model,
                                      B_B_avg,
                                      B_D1_avg,
                                      B_D2_avg,
                                      r_avg_v,
                                      v_avg_data,
                                      v_avg_model,
                                      v_B_avg,
                                      v_D1_avg,
                                      v_D2_avg,
                                      r_avg_sigma,
                                      sigma_avg_data,
                                      sigma_avg_model,
                                      sigma_B_avg,
                                      sigma_D1_avg,
                                      sigma_D2_avg,
                                      r_avg_ML,
                                      ML_avg_data,
                                      ML_avg_model,
                                      ML_B_avg,
                                      ML_D1_avg,
                                      ML_D2_avg,
                                      B_avg_err,
                                      v_avg_err,
                                      sigma_avg_err,
                                      ML_avg_err,
                                      plot_ML=plot_ML)
        
        fig.savefig(self.output_path+savename)


        tot = np.ones((int(max(Nbin_phot,Nbin_kin)),28))*np.nan

        tot[:np.size(r_avg_B),0] = r_avg_B
        tot[:np.size(B_avg_model),1] = B_avg_model
        tot[:np.size(B_B_avg),2] = B_B_avg
        tot[:np.size(B_D1_avg),3] = B_D1_avg
        tot[:np.size(B_D2_avg),4] = B_D2_avg
        tot[:np.size(B_avg_data),5] = B_avg_data
        tot[:np.size(B_avg_err),6] = B_avg_err
        tot[:np.size(r_avg_v),7] = r_avg_v
        tot[:np.size(v_avg_model),8] = v_avg_model
        tot[:np.size(v_B_avg),9] = v_B_avg
        tot[:np.size(v_D1_avg),10] = v_D1_avg
        tot[:np.size(v_D2_avg),11] = v_D2_avg
        tot[:np.size(v_avg_data),12] = v_avg_data
        tot[:np.size(v_avg_err),13] = v_avg_err
        tot[:np.size(r_avg_sigma),14] = r_avg_sigma
        tot[:np.size(sigma_avg_model),15] = sigma_avg_model
        tot[:np.size(sigma_B_avg),16] = sigma_B_avg
        tot[:np.size(sigma_D1_avg),17] = sigma_D1_avg
        tot[:np.size(sigma_D2_avg),18] = sigma_D2_avg
        tot[:np.size(sigma_avg_data),19] = sigma_avg_data
        tot[:np.size(sigma_avg_err),20] = sigma_avg_err
        tot[:np.size(r_avg_ML),21] = r_avg_ML
        tot[:np.size(ML_avg_model),22] = ML_avg_model
        tot[:np.size(ML_B_avg),23] = ML_B_avg
        tot[:np.size(ML_D1_avg),24] = ML_D1_avg
        tot[:np.size(ML_D2_avg),25] = ML_D2_avg
        tot[:np.size(ML_avg_data),26] = ML_avg_data
        tot[:np.size(ML_avg_err),27] = ML_avg_err
        np.save(self.output_path+savedata,tot)

        if show==True:
            fig.show()
            plt.show()
        return fig


    def maps_2D_specify_params(self,Best_fit_params={},savename='2D_maps.png',plot_ML=True,show=True,write=True,lim='data',vmin=None,vmax=None,close=False):
        all_params = {**Best_fit_params, **self.fixed_quantities}
        ibrightness_model,_,_,_,v_model,_,_,_,sigma_model,_,_,_,LM_ratio,_,_,_ = self.model_obj.model(all_params)
        ML_model = 1.0/LM_ratio
        #ML_model = np.log10(1.0/LM_ratio)

        fig = self.data_obj.all_maps_kpc(self.ibrightness_data,
                                    np.log10(ibrightness_model),
                                    self.v_data,
                                    v_model,
                                    self.sigma_data,
                                    sigma_model,
                                    self.ML_data,
                                    ML_model,
                                    self.v_error,
                                    self.sigma_error,
                                    self.ML_error,
                                    plot_ML=plot_ML,
                                    lim = lim,
                                    vmin=vmin,
                                    vmax=vmax)
        fig.savefig(self.output_path+savename)
        if show==True:
            fig.show()
            plt.show()
        return ibrightness_model,v_model,sigma_model,ML_model,fig

    def maps_1D_specify_params(self,Best_fit_params={},savename='1D_profiles.png',plot_ML=True,show=True,write=True,close=False):
        all_params = {**Best_fit_params , **self.fixed_quantities}
        ibrightness_model,B_B,B_D1,B_D2,v_model,v_B,v_D1,v_D2,\
        sigma_model,sigma_B,sigma_D1,sigma_D2,LM_ratio,LM_B,LM_D1,LM_D2 = self.model_obj.model(all_params)
        ML_model,ML_B,ML_D1,ML_D2 = 1.0/LM_ratio,1.0/LM_B,1.0/LM_D1,1.0/LM_D2
        #ML_model = np.log10(1.0/LM_ratio)

        minr = 0.4  # dimension of pixel is 0.5 arcsec
        maxr_kin = np.nanmax((self.x_arcsec**2+self.y_arcsec**2)**0.5*self.v_data/self.v_data)
        maxr_phot = 2*maxr_kin
        maxr_kin -= 1.5
        Nbin_phot = int((maxr_phot-minr)/1.0)
        Nbin_kin  = int((maxr_kin-minr)/1.0)

        Nbin_phot = max(Nbin_phot,7) 
        Nbin_kin  = max(Nbin_kin,7) 


        r_avg_B,B_avg_model  = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,np.log10(ibrightness_model),quantity=None,estimator='mean',
                                   min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,B_B_avg   = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,np.log10(B_B),quantity=None,estimator='mean',
                                   min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,B_D1_avg  = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,np.log10(B_D1),quantity=None,estimator='mean',
                                   min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,B_D2_avg  = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,np.log10(B_D2),quantity=None,estimator='mean',
                                   min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,B_avg_data         = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.ibrightness_data,quantity=None,estimator='mean',
                                   min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,B_avg_err          = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.ibrightness_error,quantity=None,estimator='mean',
                                   min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        
        r_avg_v,v_avg_model  = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,v_model,quantity="velocity",estimator='mean',
                                   min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=False)
        _,v_B_avg   = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,v_B,quantity="velocity",estimator='mean',
                                   min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=False)
        _,v_D1_avg  = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,v_D1,quantity="velocity",estimator='mean',
                                   min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=False)
        _,v_D2_avg  = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,v_D2,quantity="velocity",estimator='mean',
                                   min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=False)
        _,v_avg_data         = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.v_data,quantity="velocity",estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=False)
        _,v_avg_err          = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.v_error,quantity="velocity",estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=False)
        
        r_avg_sigma,sigma_avg_model = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,sigma_model,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=True)
        _,sigma_B_avg = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,sigma_B,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=True)
        _,sigma_D1_avg = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,sigma_D1,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=True)
        _,sigma_D2_avg = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,sigma_D2,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=True)
        _,sigma_avg_data            = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.sigma_data,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=True)
        _,sigma_avg_err            = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.sigma_error,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_kin,Nbins=Nbin_kin,azimuth_division=True)
        
        r_avg_ML,ML_avg_model = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,ML_model,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,ML_B_avg = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,ML_B,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,ML_D1_avg = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,ML_D1,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,ML_D2_avg = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,ML_D2,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,ML_avg_data         = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.ML_data,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)
        _,ML_avg_err          = self.result_obj.radial_profile(self.x_arcsec,self.y_arcsec,self.ML_error,quantity=None,estimator='mean',
                                                       min_r=minr,max_r=maxr_phot,Nbins=Nbin_phot,azimuth_division=True)

        #B_avg_err = np.nanmean(10**self.ibrightness_data*self.ibrightness_error*np.log(10))
        #v_avg_err = np.nanmean(self.v_error)
        #sigma_avg_err = np.nanmean(self.sigma_error)
        #ML_avg_err = np.nanmean(10**self.ML_data*self.ML_error*np.log(10))
        #ML_avg_err = np.nanmean(self.ML_error)

        fig = self.data_obj.all_1D_profile_all_outputs(
                                      self.arcsec_to_kpc,
                                      r_avg_B,
                                      B_avg_data,
                                      B_avg_model,
                                      B_B_avg,
                                      B_D1_avg,
                                      B_D2_avg,
                                      r_avg_v,
                                      v_avg_data,
                                      v_avg_model,
                                      v_B_avg,
                                      v_D1_avg,
                                      v_D2_avg,
                                      r_avg_sigma,
                                      sigma_avg_data,
                                      sigma_avg_model,
                                      sigma_B_avg,
                                      sigma_D1_avg,
                                      sigma_D2_avg,
                                      r_avg_ML,
                                      ML_avg_data,
                                      ML_avg_model,
                                      ML_B_avg,
                                      ML_D1_avg,
                                      ML_D2_avg,
                                      B_avg_err,
                                      v_avg_err,
                                      sigma_avg_err,
                                      ML_avg_err,
                                      plot_ML=plot_ML)
        
        fig.savefig(self.output_path+savename)
        if show==True:
            fig.show()
            plt.show()

        return fig

    def best_params_all_fits(self,write=True,filename='all_params.txt'):

        Best_Fit,LE_16,UE_84 = self.result_obj.best_params()
        _,LE_05,UE_95 = self.result_obj.best_params(percentile=[50,5,95])

        if os.path.exists(filename):
            f = open(filename,'a')
            f.write(self.galaxy_name+'\t')
            for nn in list(Best_Fit.keys()):
                f.write(str(np.around(Best_Fit[nn],decimals=4))+'\t')
            for nn in list(Best_Fit.keys()):
                f.write(str(np.around(LE_16[nn],decimals=4))+ '\t')
            for nn in list(Best_Fit.keys()):
                f.write(str(np.around(UE_84[nn],decimals=4))+'\t')
            for nn in list(Best_Fit.keys()):
                f.write(str(np.around(LE_05[nn],decimals=4))+'\t')
            for nn in list(Best_Fit.keys()):
                f.write(str(np.around(UE_95[nn],decimals=4))+'\t')
            f.write('\n')
        else:
            f = open(filename,'w')
            f.write('# plate-IFU\t')
            for nn in list(Best_Fit.keys()):
                f.write(nn+'\t')
            for nn in list(Best_Fit.keys()):
                f.write(nn+'_LE16 \t')
            for nn in list(Best_Fit.keys()):
                f.write(nn+'_UE84 \t')
            for nn in list(Best_Fit.keys()):
                f.write(nn+'_LE05 \t')
            for nn in list(Best_Fit.keys()):
                f.write(nn+'_UE95 \t')
            f.write('\n')

            f.write(self.galaxy_name+'\t')
            for nn in list(Best_Fit.keys()):
                f.write(str(np.around(Best_Fit[nn],decimals=4))+'\t')
            for nn in list(Best_Fit.keys()):
                f.write(str(np.around(LE_16[nn],decimals=4))+ '\t')
            for nn in list(Best_Fit.keys()):
                f.write(str(np.around(UE_84[nn],decimals=4))+'\t')
            for nn in list(Best_Fit.keys()):
                f.write(str(np.around(LE_05[nn],decimals=4))+'\t')
            for nn in list(Best_Fit.keys()):
                f.write(str(np.around(UE_95[nn],decimals=4))+'\t')
            f.write('\n')

        f.close()

    def mass_plot(self,min_r=0.1,max_r=10,Nbins=20,figname='mass_profile.png'):
        '''
        Class Method
        ----------------------------------------------------------------------
        ----------------------------------------------------------------------
        Plot mass profile for best fit model
            
        Note: this is probably working only with bulge+ inner disc+ outer disc model
        '''

        r,Mbulge,Mdisc1,Mdisc2,Mhalo = self.result_obj.mass_profile(min_r=min_r,max_r=max_r,Nbins=Nbins)
        fig = plt.figure()
        ax  = fig.add_subplot()
        ax.plot(r,Mbulge,'r',lw=3,label='Bulge')
        ax.plot(r,Mdisc1,'y',lw=3,label='Inner Disc')
        ax.plot(r,Mdisc2,'b',lw=3,label='Outer Disc')
        ax.plot(r,Mhalo,'k',lw=3,label='Halo')
        ax.set_yscale('log')
        ax.set_xlabel('r [kpc]')
        ax.set_ylabel(r'$M [M_{\odot}]$')

        fig.savefig(self.output_path+figname)

if __name__=='__main__':
    pass


import functools
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np 
import data as d 
import matplotlib.pyplot as plt
from collections import OrderedDict
'''
This is the Residual in Residual neural network structure!
'''


def make_layer(block, n_layers):
    layers = []
    for _ in range(n_layers):
        layers.append(block())
    return nn.Sequential(*layers)

# I'LL PUT THEM IN DIFFERENT CLASSES SUCH THAT IF WE WANT TO ADD SIMILAR LAYERS IS EASIER

# out1 = 52
# out2 = 12


class features_extraction(nn.Module):
    def __init__(self,img_fil,out1,kernel_size=5,bias=True):
        
        super(features_extraction,self).__init__()

        self.feat_ext = nn.Conv2d(img_fil,
                                  out1,
                                  kernel_size,
                                  stride=1,
                                  padding=2,
                                  bias=bias)
        self.p_RELU   = nn.PReLU()  # as leaky Relu but slope is learnable
        #self.p_RELU   = nn.LeakyReLU(negative_slope=0.2, inplace=True)

    def forward(self,x):
        return self.p_RELU(self.feat_ext(x))

class schrinking(nn.Module):
    def __init__(self,out1,out2,kernel_size=1,bias=True):

        super(schrinking,self).__init__()

        self.shrink = nn.Conv2d(out1,
                                out2,
                                kernel_size,
                                stride=1,
                                padding=0,
                                bias=bias)
        self.p_RELU   = nn.PReLU()
        #self.p_RELU   = nn.LeakyReLU(negative_slope=0.2, inplace=True)
    
    def forward(self,x):
        return self.p_RELU(self.shrink(x))

class mapping(nn.Module):
    def __init__(self,out2,n_layers,kernel_size=3,bias=True):
        
        super(mapping,self).__init__()

        mapping_block = OrderedDict()
        for i in range(n_layers):
            mapping_block[str(i)] = nn.Conv2d(out2,
                                              out2,
                                              kernel_size,
                                              stride=1,
                                              padding=1,
                                              bias=bias)
            mapping_block[str(i)] = nn.PReLU()
            #mapping_block[str(i)] = nn.LeakyReLU(negative_slope=0.2, inplace=True)
        
        self.mapping_block = nn.Sequential(mapping_block)

    def forward(self,x):
        return self.mapping_block(x)


class expanding(nn.Module):
    def __init__(self,out2,out1,kernel_size=1,bias=True):

        super(expanding,self).__init__()

        self.exp_layer = nn.Conv2d(out2,
                                   out1,
                                   kernel_size,
                                   stride=1,
                                   padding=0,
                                   bias=bias)
        self.p_RELU    = nn.PReLU()
        #self.p_RELU   = nn.LeakyReLU(negative_slope=0.2, inplace=True)

    def forward(self,x):
        return self.p_RELU(self.exp_layer(x))

class transp_conv(nn.Module):
    def __init__(self,out1,img_filter,kernel_size=9,scaling=4,bias=True):
        super(transp_conv,self).__init__()

        self.deconv_layer = nn.ConvTranspose2d(out1,
                                               img_filter,
                                               kernel_size,
                                               stride=2,
                                               padding=4,
                                               output_padding=1,
                                               bias=bias)
    
    def forward(self,x):
        return self.deconv_layer(x)




class FSRCNN_net(nn.Module):
    def __init__(self,img_filter,out1,out2,bias=True):
        super(FSRCNN_net,self).__init__()
        self.step1 = features_extraction(img_filter,out1)
        self.step2 = schrinking(out1,out2)
        self.step3 = mapping(out2,10)
        self.step4 = expanding(out2,out1)
        self.step5 = transp_conv(out1,2*out2)
        self.step6 = transp_conv(2*out2,img_filter)

    def forward(self,x):
        return self.step6(self.step5(self.step4(self.step3(self.step2(self.step1(x))))))



def initialize_weights(model):
    for m in model.modules():
        if isinstance(m, nn.Conv2d):
            nn.init.kaiming_normal_(m.weight.data)
            
        #elif isinstance(m, nn.ConvTranspose2d):
        #    nn.init.kaiming_normal_(m.weight.data)

if __name__=='__main__':

    device = "cuda" if torch.cuda.is_available() else "cpu"

    batch_size,img_filter,out1,out2 = 32,1,30,10

    model = FSRCNN_net(img_filter,         # filter imager grey or RGB
                       out1,               # filter for feat extraction
                       out2)               # reduced filter for mapping 
                         

    fake_img = torch.randn((batch_size,img_filter,20,20))


    HR_img = model(fake_img)

    


import numpy as np 
import matplotlib.pyplot as plt
from astropy.cosmology import FlatLambdaCDM

def grid_conversion(x,y,z):
    ''' 
    Utils Function
    ----------------------------------------------------------------------
    ----------------------------------------------------------------------
    conversion from arcsec to kpc assuming FlatLambdaCDM cosmology with H0 = 70.0 (Km/s)/Mpc and Om0 = 0.3
        
        Parameters:
            x: 1D arr float32
                x position in arcsecond
            y: 1D arr float32
                y position in arcsecond
            z: float32
                redshift
        Returns:
            x: 1D arr float32
                x position in kpc
            y: 1D arr float32
                y position in kpc
    '''

    #coordinate in radiants rad = (pi/180*3600)*arcsec
    x = (np.pi/(180*3600)) * x
    y = (np.pi/(180*3600)) * y

    H0 = 70.0           #(Km/s)/Mpc
    Om0 = 0.3
    cosmo = FlatLambdaCDM(H0=H0,Om0=Om0)

    D = cosmo.angular_diameter_distance(z)       #distance in Mpc
    D = D.value
    D = D * 1e3         #distance in Kpc
    x = D * np.tan(x)
    y = D * np.tan(y)
    D_angular = D

    return x,y

def psf_sigma_FWHM(FWHM_arcsec,z):
    ''' 
    Utils Function
    ----------------------------------------------------------------------
    ----------------------------------------------------------------------
    conversion of FWHM of the psf from arcsec to kpc 
    assuming FlatLambdaCDM cosmology with H0 = 70.0 (Km/s)/Mpc and Om0 = 0.3
        
        Parameters:
            FWHM_arcsec: float32
                FWHM of the psf in arcsec
            z: float32
                redshift
        Returns:
            sigma_kpc: float32
                sigma of the psf in kpc (assuming Gaussian)
            FWHM_kpc: float32
                FWHM of the psf in kpc (assuming Gaussian)
    '''
    H0 = 70.0           #(Km/s)/Mpc
    Om0 = 0.3
    cosmo = FlatLambdaCDM(H0=H0,Om0=Om0)
    D = cosmo.angular_diameter_distance(z)     #distance in Mpc
    D = D.value
    D = D * 1e3         #distance in Kpc

    sigma_arcsec = FWHM_arcsec/2.355

    FWHM_kpc  = np.tan((np.pi/(180*3600)) * FWHM_arcsec) * D
    sigma_kpc = np.tan((np.pi/(180*3600)) * sigma_arcsec) * D

    return(sigma_kpc,FWHM_kpc)



def nMgy_to_cgs(x):
    ''' 
    Utils Function
    ----------------------------------------------------------------------
    ----------------------------------------------------------------------
    conversion of flux from nanomaggies (SDSS) to cgs (erg/cm^2/s/Hz)
        
        Parameters:
            x: 1D arr float32
                flux in nanomaggies
        Returns:
            x_cgs: 1D arr float32
                flux in erg/s*cm^2*Hz 
    '''
    # conversion factor from nanomaggies to jansky
    conv_nMgy_Jy = 3.631e-6
    x_Jy = x * conv_nMgy_Jy

    # conversion factor from jansky to erg/s*cm^2*Hz 
    conv_Jy_cgs = 1e-23
    x_cgs = x_Jy * conv_Jy_cgs

    return x_cgs

def conversion(x,F_sun,z):
    ''' 
    Utils Function
    ----------------------------------------------------------------------
    ----------------------------------------------------------------------
    conversion of flux from cgs (erg/cm^2/s/Hz or erg/cm^2/s/AA) to brightness in  
    L_sun/pix^2, L_sun/arcsec^2 or L_sun/kpc^2 depending on the units of x
        
        Parameters:
            x: 1D arr float32
                flux in erg/s/cm^2/AA or erg/s/cm^2/Hz  
                (note that real units of x is erg/s/cm^2/Hz/(pix^2 or arcsec^2 or kpc^2 or pc^2)
            F_sun: float32
                sun flux in the same band of x at 10 pc (must have same units as x) 
        Returns:
            L_scaled: 1D arr float32
                luminosity in L_sun/pix^2, L_sun/arcsec^2 or L_sun/kpc^2 depending on the units of x
    '''


    dLE   = 3.08568e19                  # 10 pc in cm distance of the absolute magnitude
    L_sun = F_sun * 4 * np.pi * dLE**2  # Solar luminosity in that band erg/s/Hz or erg/s/AA

    H0 = 70.0                            #(Km/s)/Mpc
    Om0 = 0.3                            #Omega matter 0
    cosmo = FlatLambdaCDM(H0=H0,Om0=Om0)
    D = cosmo.luminosity_distance(z)
    D = D.value
    D = D * 1e6                   # distance in pc
    D = D * 3.08568e18            # distance in cm 
    L = 4 * np.pi * D**2 * x      # luminosity in erg/s/Hz erg/s/AA

    L_scaled = L/L_sun            # Luminosity in L_sun (per spaxel, i.e. per pixel, or per arcsec or per kpc or ...) 

    L_scaled = L_scaled           # Surface brightness in L_sun/ (pix^2  or arcsec^2 or kpc^2 or ...)


    return L_scaled

class data:

    def __init__(self,
                 x,
                 y,
                 *args, **kwargs):
        ''' 
        A class for data handling (grid manipolation) and plotting.
        WE ASSUME, SINCE WE ARE WORKING WITH MANGA THAT x and y has the same dimension
        and that they have the same number of pixels in x and y. N=J=x.shape[0]**0.5.
        x coordinates are assumed to be sorted from max to min
        ----------------------------------------------------------------------
        ----------------------------------------------------------------------                                           
            Attributes:
                    x: 1D arr float32
                        x positions of the grid in kpc or arcsec (computed from in data.py)
                    y: 1D arr float32
                        y positions of the grid in kpc or arcsec (computed from in data.py)
            Methods:
                    refined_grid_indexing(sigma_psf,N_px,resolution):
                        Compute grids starting from x,y using N_px for the convolution and possibly 
                        increasing the resolution. Compute also weights of the psf.

                    plot_2D(Z,xlabel,ylabel,zlabel):
                        2D pcolormesh plot

                    plot_3D(Z,xlabel,ylabel,zlabel):
                        3D plot

                    brightness_map(brightness,xlabel,ylabel,zlabel):
                        brightness 2D plot (useful if using default labels)

                    velocity_map(brightness,xlabel,ylabel,zlabel):
                        velocity 2D plot (useful if using default labels)

                    dispersion_map(brightness,xlabel,ylabel,zlabel):
                        dispersion 2D plot (useful if using default labels)

                    ML_map(brightness,xlabel,ylabel,zlabel):
                        Mass to light 2D plot (useful if using default labels)

                    brightness_3D(brightness,xlabel,ylabel,zlabel):
                        brightness 3D plot (useful if using default labels)

                    velocity_3D(brightness,xlabel,ylabel,zlabel):
                        velocity 3D plot (useful if using default labels)

                    dispersion_3D(brightness,xlabel,ylabel,zlabel):
                        dispersion 3D plot (useful if using default labels)

                    ML_3D(brightness,xlabel,ylabel,zlabel):
                        Mass to light 3D plot (useful if using default labels)

                    all_maps_kpc(brightness_data,brightness_model,v_data,v_model,sigma_data,sigma_model,ML_data,ML_model,
                                 brightness_error,v_error,sigma_error,ML_error,plot_ML=True,lim='data',vmin=None,vmax=None):
                        Combine 2D maps of 3X3 or 3X4 plot with labels in kpc

                    all_maps_arcsec(brightness_data,brightness_model,v_data,v_model,sigma_data,sigma_model,ML_data,ML_model,
                                    v_error,sigma_error,ML_error,plot_ML=True,lim='data',vmin=None,vmax=None):
                        Combine 2D maps of 3X3 or 3X4 plot with labels in arcsec

                    all_1D_profile(arcsec_to_kpc,r_avg_brightness,brightness_avg_data,brightness_avg_model,r_avg_v,v_avg_data,
                                    v_avg_model,r_avg_sigma,sigma_avg_data,sigma_avg_model,r_avg_ML,ML_avg_data,ML_avg_model,
                                    brightness_avg_error,v_avg_error,sigma_avg_error,ML_avg_error,plot_ML=True):
                        Combine azimuthally averaged profiles on 1X3 or 1X4 plot with labels both in arcsec and kpc

                    all_1D_profile_all_outputs(arcsec_to_kpc,r_avg_brightness,brightness_avg_data,brightness_avg_model,
                                                   B_B_avg,B_D1_avg,B_D2_avg,r_avg_v,v_avg_data,v_avg_model,v_B_avg,v_D1_avg,v_D2_avg,
                                                   r_avg_sigma,sigma_avg_data,sigma_avg_model,sigma_B_avg,sigma_D1_avg,sigma_D2_avg,r_avg_ML,
                                                   ML_avg_data,ML_avg_model,ML_B_avg,ML_D1_avg,ML_D2_avg,brightness_avg_error,v_avg_error,
                                                   sigma_avg_error,ML_avg_error,plot_ML=True):
                        Combine azimuthally averaged profiles (of each component) on 1X3 or 1X4 plot with labels both in arcsec and kpc
        Note:
            WE ASSUME, SINCE WE ARE WORKING WITH MANGA THAT x and y has the same dimension
            and that they have the same number of pixels in x and y. N=J=x.shape[0]**0.5.
            x coordinates are assumed to be sorted from max to min
        '''

        self.x          = x
        self.y          = y

        #min and max of the grid
        self.xmin, self.xmax = np.min(x), np.max(x)
        self.ymin, self.ymax = np.min(y), np.max(y)


        self.J,self.N = int(self.x.shape[0]**0.5),int(self.x.shape[0]**0.5)
        #vectors with only the numbers without repetition
        #x_true has dimension N, while y_true has dimension J
        self.x_true = self.x[0:self.N]  #x[0 : J*N : J]
        self.y_true = self.y[0:self.J*self.N :self.J] 

        #dimension of the grid, dx size N-1 and dy size J-1. The grid is not supposed to be const
        self.dx = abs(self.x_true[0]-self.x_true[1])
        self.dy = self.dx


       

#    def refined_grid(self,sigma_psf=0.2664348539150563,N_px=10):
#
#        self.x_true = np.linspace(self.xmin,self.xmax,self.J)
#        self.y_true = np.linspace(self.ymin,self.ymax,self.N)
#        
#        self.x_true = self.x_true[::-1]
#        x,y = np.meshgrid(self.x_true,self.y_true,indexing = 'ij')
#        x,y = x.ravel(),y.ravel()
#        
#        self.dx = abs(self.x_true[1]-self.x_true[0])
#        
#        self.k = 2*N_px+1
#        self.K = int(self.k*self.k)
#        
#        self.new_true_x = np.linspace(-N_px*self.dx,N_px*self.dx,2*N_px+1)
#        self.new_true_y = np.linspace(-N_px*self.dx,N_px*self.dx,2*N_px+1)
#
#        self.new_true_x = self.new_true_x[::-1]
#
#        print('Number of points for the convolution is {}'.format(self.K))
#
#        x_rand = np.zeros((self.x.shape[0],self.K))
#        y_rand = np.zeros((self.x.shape[0],self.K))
#
#        x_tmp_g,y_tmp_g = np.meshgrid(self.new_true_x,self.new_true_y,indexing='ij')
#        x_tmp_g,y_tmp_g = x_tmp_g.ravel(),y_tmp_g.ravel()
#        r = np.sqrt(x_tmp_g**2+y_tmp_g**2)
#
#        for i in range(self.x.shape[0]):
#            x_rand[i,:] = x_tmp_g+x[i]
#            y_rand[i,:] = y_tmp_g+y[i]
#
#        x_rand,y_rand = x_rand.ravel(),y_rand.ravel()
#        psf    = np.exp(-r*r/(2*sigma_psf*sigma_psf))
#
#        return(x_rand,y_rand,psf,self.K)


    def refined_grid_indexing(self,sigma_psf=0.2664348539150563,N_px=10,resolution=1):
        ''' 
        Class Method
        ----------------------------------------------------------------------
        ----------------------------------------------------------------------
        Compute grids starting from x,y using N_px for the convolution and possibly 
        increasing the resolution. Compute also weights of the psf.
            
            Parameters:
                sigma_psf: float32
                    sigma of the psf in kpc assuming Gaussian
                N_px: int16
                    number of pixels to consider in the convolution
                resolution: int16
                    use pixel (resolution=1) or subpixel (resolution>1) resolution for
                    psf convolution

            Returns:
                new_grid_x_flat: 1D arr float32
                    new grid in x. It covers the same points as self.x but it has 
                    N_px pixels more in each direction.
                new_grid_y_flat: 1D arr float32
                    new grid in y. It covers the same points as self.y but it has 
                    N_px pixels more in each direction.
                psf: 1D arr float32
                    Weights of the psf. It is a vector of dimension K. 
                    Since the dimension of the pixels is constant it is not necessary to compute
                    weights for all the pixels.
                K: float32
                    Number of pixels to consider for the psf convolution. The value at point x[i],y[i]
                    is the weighted average over K points around x[i],y[i]
                indexing: 1D arr int16
                    Vector of K x N x J points. The first indexing[:K] numbers are the indexes of the points
                    to be considere for the psf. It means that for example:
                    rho(x[0],y[0]) is the "average" of psf*rho(new_grid_x_flat[:K],new_grid_y_flat[:K]) 
        Note:
            x coordinates are from max to min
        '''
        print('remember that x coordinates are assumed to be from max to min')

        if not isinstance(N_px, int):
            print('error N_px must be integer')
            exit()
        if not isinstance(resolution,int):
            print('error resolution must be integer')
            exit()

        if N_px == 0:
            self.x_true = np.linspace(self.xmin,self.xmax,self.J)
            self.y_true = np.linspace(self.ymin,self.ymax,self.N)
            ind_min_x     = np.argmin(self.x_true*self.x_true)
            ind_min_y     = np.argmin(self.y_true*self.y_true)
            self.x_true = self.x_true - self.x_true[ind_min_x] 
            self.y_true = self.y_true - self.y_true[ind_min_y]
            self.x_true =self.x_true[::-1]
            new_grid_x,new_grid_y = np.meshgrid(self.x_true,self.y_true)

            return new_grid_x.ravel(),new_grid_y.ravel(),np.ones(1),1,np.arange(self.x.shape[0],dtype=np.int32)

        self.x_true = np.linspace(self.xmin,self.xmax,self.J)
        self.y_true = np.linspace(self.ymin,self.ymax,self.N)

        self.dx = abs(self.x_true[1]-self.x_true[0])
        hr_vec = np.linspace(0.,self.dx,resolution+1)
        #self.dx = np.round(self.dx,decimals=4)

        new_true_x_tmp = np.linspace(self.xmin-N_px*self.dx,self.xmax+N_px*self.dx,self.J+2*N_px)
        new_true_y_tmp = np.linspace(self.ymin-N_px*self.dx,self.ymax+N_px*self.dx,self.N+2*N_px)

        # Make sure it is in zero
        ind_min_new_x = np.argmin(new_true_x_tmp*new_true_x_tmp)
        ind_min_new_y = np.argmin(new_true_y_tmp*new_true_y_tmp)
        ind_min_x     = np.argmin(self.x_true*self.x_true)
        ind_min_y     = np.argmin(self.y_true*self.y_true)
        new_true_x_tmp= new_true_x_tmp-new_true_x_tmp[ind_min_new_x]
        new_true_y_tmp= new_true_y_tmp-new_true_y_tmp[ind_min_new_y] 
        self.x_true = self.x_true - self.x_true[ind_min_x] 
        self.y_true = self.y_true - self.y_true[ind_min_y]

        self.new_true_x = np.array([])
        self.new_true_y = np.array([])
        if resolution < 1:
            print('Error: resolution lower than pixel size is not allowed ')
            exit()
        elif resolution >1:
            hr_vec = hr_vec[:-1]
            for i in range(new_true_x_tmp.shape[0]):
                self.new_true_x = np.append(self.new_true_x,hr_vec+new_true_x_tmp[i])
            for i in range(new_true_y_tmp.shape[0]):
                self.new_true_y = np.append(self.new_true_y,hr_vec+new_true_y_tmp[i])
        else: 
            self.new_true_x = np.copy(new_true_x_tmp)
            self.new_true_y = np.copy(new_true_y_tmp)
        
        # for this case coordinate are from max to min in the x axis
        self.x_true =self.x_true[::-1]
        self.new_true_x = self.new_true_x[::-1]
        new_grid_x,new_grid_y = np.meshgrid(self.new_true_x,self.new_true_y)#,indexing = 'ij')
        new_N,new_J = new_grid_x.shape[0],new_grid_x.shape[1]
        
        x = new_grid_x[N_px:-N_px,N_px:-N_px]
        y = new_grid_y[N_px:-N_px,N_px:-N_px]
        
        new_grid_x_flat,new_grid_y_flat = new_grid_x.ravel(),new_grid_y.ravel()


        # questa posso pensare di prenderla da un sottoinsieme dell'altra
        #x,y = np.meshgrid(self.x_true,self.y_true)#,indexing = 'ij')
        x,y = x.ravel(),y.ravel()

        indexing = np.zeros((x.shape[0],(2*N_px*resolution+1)*(2*N_px*resolution+1)),dtype=np.int32)

        integer_index = np.array([i for i in range(-N_px*resolution,resolution*N_px+1)])
        for i in range(x.shape[0]):
            # Per questioni di confronto numerico faccio così:
            #       - scelgo i punti che distano (N_px+1)*dx dal centro
            #       - li ordino per distanza dal centro 
            #       - scelgo i (2N_px+1)*(2N_px+1) più vicini
            #       - sorto gli indici dal più piccolo al più grande
            x_tmp,y_tmp = x[i],y[i]
            r_mat = (x_tmp-new_grid_x)**2+(y_tmp-new_grid_y)**2
            min_index = np.unravel_index(np.argmin(r_mat, axis=None), r_mat.shape)
            x_index = integer_index+min_index[0]
            y_index = integer_index+min_index[1]
            x_index,y_index = np.meshgrid(x_index,y_index,indexing='ij')
            x_index,y_index = x_index.ravel(),y_index.ravel()

            flat_index = y_index + x_index*new_J
            flat_index = np.int32(flat_index)
            indexing[i,:] = flat_index
        
        self.K = indexing.shape[1]
        # compute psf weight
        x_tmp,y_tmp = x[0],y[0]
        ind_tmp = indexing[0,:]
        r = ((new_grid_x_flat[ind_tmp]-x_tmp)**2+(new_grid_y_flat[ind_tmp]-y_tmp)**2)**0.5    

        psf = np.exp(-r*r/2./(sigma_psf**2))
        
        indexing = indexing.ravel()
        print('Number of points for the convolution is {}'.format(self.K))

        return new_grid_x_flat,new_grid_y_flat,psf,self.K,indexing

    def plot_2D(self,Z,xlabel,ylabel,zlabel):
        X = self.x.reshape((self.N,self.J))
        Y = self.y.reshape((self.N,self.J))

        fig = plt.figure()
        ax = fig.add_subplot()    
        ax.set_aspect('equal', 'box')  
        pmesh = ax.pcolormesh(X,Y,Z.T,cmap = 'viridis',shading='nearest')
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_xlim(self.xmin,self.xmax)
        ax.set_ylim(self.ymin,self.ymax)
        cbar = plt.colorbar(pmesh)
        cbar.ax.set_ylabel(zlabel)

        return (fig,ax,pmesh,cbar)

    def plot_3D(self,Z,xlabel,ylabel,zlabel):
        fig = plt.figure()
        ax =fig.add_subplot(projection = '3d')
        ax.scatter(self.x,self.y,Z,c = 'black',s = 1.5,marker = '.')
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_zlabel(zlabel)
        ax.set_xlim(self.xmin,self.xmax)
        ax.set_ylim(self.ymin,self.ymax)

        return (fix,ax)

    def brightness_map(self,brightness,xlabel='x [kpc]',ylabel='y [kpc]',zlabel= r'$log_{10}(B)$ $[L_{\odot}/kpc^2]$'):
        if brightness.shape == (self.N,self.J):
            B_matrix = np.copy(brightness)
        else:
            B_matrix = brightness.reshape((self.N,self.J))
        fig,ax,pmesh,cbar = self.plot_2D(B_matrix,xlabel,ylabel,zlabel)
        return (fig,ax,pmesh,cbar)

    def velocity_map(self,v_los,xlabel='x [kpc]',ylabel='y [kpc]',zlabel= r'$v_{los}$ $[km/s]$'):
        if v_los.shape == (self.N,self.J):
            v_matrix = np.copy(v_los)
        else:
            v_matrix = v_los.reshape((self.N,self.J))
        fig,ax,pmesh,cbar = self.plot_2D(v_matrix,xlabel,ylabel,zlabel)
        return(fig,ax,pmesh,cbar)

    def dispersion_map(self,sigma_los,xlabel='x [kpc]',ylabel='y [kpc]',zlabel= r'$\sigma_{los}$ $[km/s]$'):
        if sigma_los.shape == (self.N,self.J):
            sigma_matrix = np.copy(sigma_los)
        else:
            sigma_matrix = sigma_los.reshape((self.N,self.J))
        fig,ax,pmesh,cbar = self.plot_2D(sigma_matrix,xlabel,ylabel,zlabel)
        return(fig,ax,pmesh,cbar)
    
    def ML_map(self,ML_ratio,xlabel='x [kpc]',ylabel='y [kpc]',zlabel= r'$M/L$ $[M_{\odot}/L_{\odot}]$'):
        if ML_ratio.shape == (self.N,self.J):
            ML_matrix = np.copy(ML_ratio)
        else:
            ML_matrix = ML_ratio.reshape((self.N,self.J))
        fig,ax,pmesh,cbar = self.plot_2D(ML_matrix,xlabel,ylabel,zlabel)
        return(fig,ax,pmesh,cbar)


    def brightness_3D(self,brightness,xlabel='x [kpc]',ylabel='y [kpc]',zlabel= r'$log_{10}(B)$ $[L_{\odot}/kpc^2]$'):
        fig,ax = self.plot_3D(brightness,xlabel,ylabel,zlabel)
        return(fig,ax)

    def velocity_3D(self,v_los,xlabel='x [kpc]',ylabel='y [kpc]',zlabel= r'$v_{los}$ $[km/s]$'):
        fig,ax = self.plot_3D(v_los,xlabel,ylabel,zlabel)
        return(fig,ax)

    def dispersion_3D(self,sigma_los,xlabel='x [kpc]',ylabel='y [kpc]',zlabel= r'$\sigma_{los}$ $[km/s]$'):
        fig,ax = self.plot_3D(sigma_los,xlabel,ylabel,zlabel)
        return(fig,ax)

    def ML_ratio_3D(self,ML_ratio,xlabel='x [kpc]',ylabel='y [kpc]',zlabel=  r'$M/L$ $[M_{\odot}/L_{\odot}]$'):
        fig,ax = self.plot_3D(ML_ratio,xlabel,ylabel,zlabel)
        return(fig,ax)

    def all_maps_kpc(self,
                     brightness_data,
                     brightness_model,
                     v_data,
                     v_model,
                     sigma_data,
                     sigma_model,
                     ML_data,
                     ML_model,
                     brightness_error,
                     v_error,
                     sigma_error,
                     ML_error,
                     plot_ML=True,
                     lim='data',
                     vmin=None,
                     vmax=None):
        
        if brightness_data.shape == (self.N,self.J):
            B_matrix_data = np.copy(brightness_data)
        else:
            B_matrix_data = brightness_data.reshape((self.N,self.J))
        if brightness_model.shape == (self.N,self.J):
            B_matrix_model = np.copy(brightness_model)
        else:
            B_matrix_model = brightness_model.reshape((self.N,self.J))
        if brightness_error.shape == (self.N,self.J):
            B_matrix_err = np.copy(brightness_error)    
        else:
            B_matrix_err = brightness_error.reshape((self.N,self.J))
        if v_data.shape == (self.N,self.J):
            v_matrix_data = np.copy(v_data)
        else:
            v_matrix_data = v_data.reshape((self.N,self.J))
        if v_model.shape == (self.N,self.J):
            v_matrix_model = np.copy(v_model)
        else:
            v_matrix_model = v_model.reshape((self.N,self.J))
        if v_error.shape == (self.N,self.J):
            v_matrix_err = np.copy(v_error)    
        else:
            v_matrix_err = v_error.reshape((self.N,self.J))
        if sigma_data.shape == (self.N,self.J):
            sigma_matrix_data = np.copy(sigma_data)
        else:
            sigma_matrix_data = sigma_data.reshape((self.N,self.J))
        if sigma_model.shape == (self.N,self.J):
            sigma_matrix_model = np.copy(sigma_model)
        else:
            sigma_matrix_model = sigma_model.reshape((self.N,self.J))
        if sigma_error.shape == (self.N,self.J):
            sigma_matrix_err = np.copy(sigma_error)    
        else:
            sigma_matrix_err = sigma_error.reshape((self.N,self.J))    
        if ML_data.shape == (self.N,self.J):
            ML_matrix_data = np.copy(ML_data)
        else:
            ML_matrix_data = ML_data.reshape((self.N,self.J))
        if ML_model.shape == (self.N,self.J):
            ML_matrix_model = np.copy(ML_model)
        else:
            ML_matrix_model = ML_model.reshape((self.N,self.J))
        if ML_error.shape == (self.N,self.J):
            ML_matrix_err = np.copy(ML_error)    
        else:
            ML_matrix_err = ML_error.reshape((self.N,self.J))  

        if lim=='data':
            vmin = [np.nanmin(brightness_data),np.nanmin(v_data),np.nanmin(sigma_data),np.nanmin(ML_data)]
            vmax = [np.nanmax(brightness_data),np.nanmax(v_data),np.nanmax(sigma_data),np.nanmax(ML_data)]
        elif lim=='model':
            vmin = [np.nanmin(brightness_model),np.nanmin(v_model),np.nanmin(sigma_model),np.nanmin(ML_model)]
            vmax = [np.nanmax(brightness_model),np.nanmax(v_model),np.nanmax(sigma_model),np.nanmax(ML_model)]
        elif lim=='user':
            if vmin==None or vmax==None:
                print('error: must specify limits')
        else:
            print('lim can be data or model or user')

        X = self.x.reshape((self.N,self.J))
        Y = self.y.reshape((self.N,self.J))

        fig = plt.figure()
        fig.set_figheight(20)
        fig.set_figwidth(25)
        
        if plot_ML == True: 
            ax1  = fig.add_subplot(3,4,1) 
            ax2  = fig.add_subplot(3,4,2) 
            ax3  = fig.add_subplot(3,4,3) 
            ax4  = fig.add_subplot(3,4,4) 
            ax5  = fig.add_subplot(3,4,5) 
            ax6  = fig.add_subplot(3,4,6) 
            ax7  = fig.add_subplot(3,4,7) 
            ax8  = fig.add_subplot(3,4,8) 
            ax9  = fig.add_subplot(3,4,9)
            ax10 = fig.add_subplot(3,4,10) 
            ax11 = fig.add_subplot(3,4,11) 
            ax12 = fig.add_subplot(3,4,12) 
            ax1.set_xlabel('x [kpc]'),ax1.set_ylabel('y [kpc]')
            ax2.set_xlabel('x [kpc]'),ax2.set_ylabel('y [kpc]')
            ax3.set_xlabel('x [kpc]'),ax3.set_ylabel('y [kpc]')
            ax4.set_xlabel('x [kpc]'),ax4.set_ylabel('y [kpc]')
            ax5.set_xlabel('x [kpc]'),ax5.set_ylabel('y [kpc]')
            ax6.set_xlabel('x [kpc]'),ax6.set_ylabel('y [kpc]')
            ax7.set_xlabel('x [kpc]'),ax7.set_ylabel('y [kpc]')
            ax8.set_xlabel('x [kpc]'),ax8.set_ylabel('y [kpc]')
            ax9.set_xlabel('x [kpc]'),ax9.set_ylabel('y [kpc]')
            ax10.set_xlabel('x [kpc]'),ax10.set_ylabel('y [kpc]')
            ax11.set_xlabel('x [kpc]'),ax11.set_ylabel('y [kpc]')
            ax12.set_xlabel('x [kpc]'),ax12.set_ylabel('y [kpc]')
            ax1.set_xlim(self.xmin,self.xmax),ax1.set_ylim(self.ymin,self.ymax)
            ax2.set_xlim(self.xmin,self.xmax),ax2.set_ylim(self.ymin,self.ymax)
            ax3.set_xlim(self.xmin,self.xmax),ax3.set_ylim(self.ymin,self.ymax)
            ax4.set_xlim(self.xmin,self.xmax),ax4.set_ylim(self.ymin,self.ymax)
            ax5.set_xlim(self.xmin,self.xmax),ax5.set_ylim(self.ymin,self.ymax)
            ax6.set_xlim(self.xmin,self.xmax),ax6.set_ylim(self.ymin,self.ymax)
            ax7.set_xlim(self.xmin,self.xmax),ax7.set_ylim(self.ymin,self.ymax)
            ax8.set_xlim(self.xmin,self.xmax),ax8.set_ylim(self.ymin,self.ymax)
            ax9.set_xlim(self.xmin,self.xmax),ax9.set_ylim(self.ymin,self.ymax)
            ax10.set_xlim(self.xmin,self.xmax),ax10.set_ylim(self.ymin,self.ymax)
            ax11.set_xlim(self.xmin,self.xmax),ax11.set_ylim(self.ymin,self.ymax)
            ax12.set_xlim(self.xmin,self.xmax),ax12.set_ylim(self.ymin,self.ymax)
            ax1.set_aspect('equal', 'box')  
            ax2.set_aspect('equal', 'box')  
            ax3.set_aspect('equal', 'box')  
            ax4.set_aspect('equal', 'box')  
            ax5.set_aspect('equal', 'box')  
            ax6.set_aspect('equal', 'box')  
            ax7.set_aspect('equal', 'box')  
            ax8.set_aspect('equal', 'box')  
            ax9.set_aspect('equal', 'box')  
            ax10.set_aspect('equal', 'box')  
            ax11.set_aspect('equal', 'box')  
            ax12.set_aspect('equal', 'box')  

            # BRIGHTNESS
            pmesh1 = ax1.pcolormesh(X,Y,B_matrix_data,cmap = 'viridis',shading='nearest',vmin=vmin[0],vmax=vmax[0])
            cbar1  = plt.colorbar(pmesh1,ax=ax1,label=r'$log_{10}(B)$ $[L_{\odot}/arcsec^2]$')
            #B_lim  = pmesh1.get_clim()
            pmesh5 = ax5.pcolormesh(X,Y,B_matrix_model,cmap = 'viridis',shading='nearest',vmin=vmin[0],vmax=vmax[0])
            cbar5  = plt.colorbar(pmesh5,ax=ax5,label=r'$log_{10}(B)$ $[L_{\odot}/arcsec^2]$' )
            #pmesh5.set_clim(B_lim)
            #pmesh9 = ax9.pcolormesh(X,Y,(10**B_matrix_data-10**B_matrix_model)/10**B_matrix_data,cmap = 'viridis',shading='nearest')
            pmesh9 = ax9.pcolormesh(X,Y,(10**B_matrix_data-10**B_matrix_model)/B_matrix_err,cmap = 'viridis',shading='nearest')
            #cbar9  = plt.colorbar(pmesh9,ax=ax9,label=r'$(B_{data}-B_{model})$ / $B_{data}$')
            cbar9  = plt.colorbar(pmesh9,ax=ax9,label=r'$(B_{data}-B_{model})$ / $\delta B$')
            #pmesh9.set_clim((-1.,1.))
            pmesh9.set_clim((-10.,10.))

            # VELOCITY
            pmesh2 = ax2.pcolormesh(X,Y,v_matrix_data,cmap = 'viridis',shading='nearest',vmin=vmin[1],vmax=vmax[1])
            cbar2  = plt.colorbar(pmesh2,ax=ax2,label=r'$v_{los}$ $[km/s]$')
            #v_lim  = pmesh2.get_clim()
            pmesh6 = ax6.pcolormesh(X,Y,v_matrix_model,cmap = 'viridis',shading='nearest',vmin=vmin[1],vmax=vmax[1])
            cbar6  = plt.colorbar(pmesh6,ax=ax6,label=r'$v_{los}$ $[km/s]$')
            #pmesh6.set_clim((v_lim))
            pmesh10 = ax10.pcolormesh(X,Y,(v_matrix_data-v_matrix_model)/v_matrix_err,cmap = 'viridis',shading='nearest')
            cbar10  = plt.colorbar(pmesh10,ax=ax10,label=r'$(v_{data}-v_{model})$ / $\delta v$')
            pmesh10.set_clim((-10.,10.))

            # SIGMA
            pmesh3 = ax3.pcolormesh(X,Y,sigma_matrix_data,cmap = 'viridis',shading='nearest',vmin=vmin[2],vmax=vmax[2])
            cbar3  = plt.colorbar(pmesh3,ax=ax3,label=r'$\sigma_{los}$ $[km/s]$')
            #sigma_lim  = pmesh3.get_clim()
            pmesh7 = ax7.pcolormesh(X,Y,sigma_matrix_model,cmap = 'viridis',shading='nearest',vmin=vmin[2],vmax=vmax[2])
            cbar7  = plt.colorbar(pmesh7,ax=ax7,label=r'$\sigma_{los}$ $[km/s]$')
            #pmesh7.set_clim((sigma_lim))
            pmesh11 = ax11.pcolormesh(X,Y,(sigma_matrix_data-sigma_matrix_model)/sigma_matrix_err,cmap = 'viridis',shading='nearest')
            cbar11  = plt.colorbar(pmesh11,ax=ax11,label=r'$(\sigma_{data}-\sigma_{model})$ / $\delta \sigma$')
            pmesh11.set_clim((-10.,10.))

            # ML ratio
            pmesh4 = ax4.pcolormesh(X,Y,ML_matrix_data,cmap = 'viridis',shading='nearest',vmin=vmin[3],vmax=vmax[3])
            cbar4  = plt.colorbar(pmesh4,ax=ax4,label=r'$ML$ $[M_{\odot}/L_{\odot}]$')
            #ML_lim  = pmesh4.get_clim()
            pmesh8 = ax8.pcolormesh(X,Y,ML_matrix_model,cmap = 'viridis',shading='nearest',vmin=vmin[3],vmax=vmax[3])
            cbar8  = plt.colorbar(pmesh8,ax=ax8,label=r'$ML$ $[M_{\odot}/L_{\odot}]$')
            #pmesh8.set_clim((ML_lim))
            pmesh12 = ax12.pcolormesh(X,Y,(ML_matrix_data-ML_matrix_model)/ML_matrix_err,cmap = 'viridis',shading='nearest')
            cbar12  = plt.colorbar(pmesh12,ax=ax12,label=r'$(ML_{data}-ML_{model})$ / $\delta ML$')
            pmesh12.set_clim((-10.,10.))

            plt.subplots_adjust(top=0.99,
                                bottom=0.07,
                                left=0.005,
                                right=0.98,
                                hspace=0.305,
                                wspace=0.18)

        else: 
            ax1  = fig.add_subplot(3,3,1)
            ax2  = fig.add_subplot(3,3,2)
            ax3  = fig.add_subplot(3,3,3)
            ax4  = fig.add_subplot(3,3,4)
            ax5  = fig.add_subplot(3,3,5)
            ax6  = fig.add_subplot(3,3,6)
            ax7  = fig.add_subplot(3,3,7)
            ax8  = fig.add_subplot(3,3,8)
            ax9  = fig.add_subplot(3,3,9)
        
            ax1.set_xlabel('x [kpc]'),ax1.set_ylabel('y [kpc]')
            ax2.set_xlabel('x [kpc]'),ax2.set_ylabel('y [kpc]')
            ax3.set_xlabel('x [kpc]'),ax3.set_ylabel('y [kpc]')
            ax4.set_xlabel('x [kpc]'),ax4.set_ylabel('y [kpc]')
            ax5.set_xlabel('x [kpc]'),ax5.set_ylabel('y [kpc]')
            ax6.set_xlabel('x [kpc]'),ax6.set_ylabel('y [kpc]')
            ax7.set_xlabel('x [kpc]'),ax7.set_ylabel('y [kpc]')
            ax8.set_xlabel('x [kpc]'),ax8.set_ylabel('y [kpc]')
            ax9.set_xlabel('x [kpc]'),ax9.set_ylabel('y [kpc]')
            ax1.set_xlim(self.xmin,self.xmax),ax1.set_ylim(self.ymin,self.ymax)
            ax2.set_xlim(self.xmin,self.xmax),ax2.set_ylim(self.ymin,self.ymax)
            ax3.set_xlim(self.xmin,self.xmax),ax3.set_ylim(self.ymin,self.ymax)
            ax4.set_xlim(self.xmin,self.xmax),ax4.set_ylim(self.ymin,self.ymax)
            ax5.set_xlim(self.xmin,self.xmax),ax5.set_ylim(self.ymin,self.ymax)
            ax6.set_xlim(self.xmin,self.xmax),ax6.set_ylim(self.ymin,self.ymax)
            ax7.set_xlim(self.xmin,self.xmax),ax7.set_ylim(self.ymin,self.ymax)
            ax8.set_xlim(self.xmin,self.xmax),ax8.set_ylim(self.ymin,self.ymax)
            ax9.set_xlim(self.xmin,self.xmax),ax9.set_ylim(self.ymin,self.ymax)
            ax1.set_aspect('equal', 'box')  
            ax2.set_aspect('equal', 'box')  
            ax3.set_aspect('equal', 'box')  
            ax4.set_aspect('equal', 'box')  
            ax5.set_aspect('equal', 'box')  
            ax6.set_aspect('equal', 'box')  
            ax7.set_aspect('equal', 'box')  
            ax8.set_aspect('equal', 'box')  
            ax9.set_aspect('equal', 'box') 
            # BRIGHTNESS
            pmesh1 = ax1.pcolormesh(X,Y,B_matrix_data,cmap = 'viridis',shading='nearest',vmin=vmin[0],vmax=vmax[0])
            cbar1  = plt.colorbar(pmesh1,ax=ax1,label=r'$log_{10}(B)$ $[L_{\odot}/arcsec^2]$')
            #B_lim  = pmesh1.get_clim()
            pmesh4 = ax4.pcolormesh(X,Y,B_matrix_model,cmap = 'viridis',shading='nearest',vmin=vmin[0],vmax=vmax[0])
            cbar4  = plt.colorbar(pmesh4,ax=ax4,label=r'$log_{10}(B)$ $[L_{\odot}/arcsec^2]$')
            #pmesh4.set_clim(B_lim)
            pmesh7 = ax7.pcolormesh(X,Y,(10**B_matrix_data-10**B_matrix_model)/10**B_matrix_data,cmap = 'viridis',shading='nearest')
            cbar7  = plt.colorbar(pmesh7,ax=ax7,label=r'$(B_{data}-B_{model})$ / $B_{data}$')
            pmesh7.set_clim((-1.,1.))

            # VELOCITY
            pmesh2 = ax2.pcolormesh(X,Y,v_matrix_data,cmap = 'viridis',shading='nearest',vmin=vmin[1],vmax=vmax[1])
            cbar2  = plt.colorbar(pmesh2,ax=ax2,label=r'$v_{los}$ $[km/s]$')
            #v_lim  = pmesh2.get_clim()
            pmesh5 = ax5.pcolormesh(X,Y,v_matrix_model,cmap = 'viridis',shading='nearest',vmin=vmin[1],vmax=vmax[1])
            cbar5  = plt.colorbar(pmesh5,ax=ax5,label=r'$v_{los}$ $[km/s]$')
            #pmesh5.set_clim((v_lim))
            pmesh8 = ax8.pcolormesh(X,Y,(v_matrix_data-v_matrix_model)/v_matrix_err,cmap = 'viridis',shading='nearest')
            cbar8  = plt.colorbar(pmesh8,ax=ax8,label=r'$(v_{data}-v_{model})$ / $\delta v$')
            pmesh8.set_clim((-10.,10.))

            # SIGMA
            pmesh3 = ax3.pcolormesh(X,Y,sigma_matrix_data,cmap = 'viridis',shading='nearest',vmin=vmin[2],vmax=vmax[2])
            cbar3  = plt.colorbar(pmesh3,ax=ax3,label=r'$\sigma_{los}$ $[km/s]$')
            #sigma_lim  = pmesh3.get_clim()
            pmesh6 = ax6.pcolormesh(X,Y,sigma_matrix_model,cmap = 'viridis',shading='nearest',vmin=vmin[2],vmax=vmax[2])
            cbar6  = plt.colorbar(pmesh6,ax=ax6,label=r'$\sigma_{los}$ $[km/s]$')
            #pmesh6.set_clim((sigma_lim))
            pmesh9 = ax9.pcolormesh(X,Y,(sigma_matrix_data-sigma_matrix_model)/sigma_matrix_err,cmap = 'viridis',shading='nearest')
            cbar9  = plt.colorbar(pmesh9,ax=ax9,label=r'$(\sigma_{data}-\sigma_{model})$ / $\delta \sigma$')
            pmesh9.set_clim((-10.,10.))
        
            plt.subplots_adjust(top=0.98,
                                bottom=0.075,
                                left=0.1,
                                right=0.905,
                                hspace=0.29,
                                wspace=0.2)


        return fig


    def all_maps_arcsec(self,
                     brightness_data,
                     brightness_model,
                     v_data,
                     v_model,
                     sigma_data,
                     sigma_model,
                     ML_data,
                     ML_model,
                     v_error,
                     sigma_error,
                     ML_error,
                     plot_ML=True,
                     lim='data',
                     vmin=None,
                     vmax=None):
        
        if brightness_data.shape == (self.N,self.J):
            B_matrix_data = np.copy(brightness_data)
        else:
            B_matrix_data = brightness_data.reshape((self.N,self.J))
        if brightness_model.shape == (self.N,self.J):
            B_matrix_model = np.copy(brightness_model)
        else:
            B_matrix_model = brightness_model.reshape((self.N,self.J))
        if v_data.shape == (self.N,self.J):
            v_matrix_data = np.copy(v_data)
        else:
            v_matrix_data = v_data.reshape((self.N,self.J))
        if v_model.shape == (self.N,self.J):
            v_matrix_model = np.copy(v_model)
        else:
            v_matrix_model = v_model.reshape((self.N,self.J))
        if v_error.shape == (self.N,self.J):
            v_matrix_err = np.copy(v_error)    
        else:
            v_matrix_err = v_error.reshape((self.N,self.J))
        if sigma_data.shape == (self.N,self.J):
            sigma_matrix_data = np.copy(sigma_data)
        else:
            sigma_matrix_data = sigma_data.reshape((self.N,self.J))
        if sigma_model.shape == (self.N,self.J):
            sigma_matrix_model = np.copy(sigma_model)
        else:
            sigma_matrix_model = sigma_model.reshape((self.N,self.J))
        if sigma_error.shape == (self.N,self.J):
            sigma_matrix_err = np.copy(sigma_error)    
        else:
            sigma_matrix_err = sigma_error.reshape((self.N,self.J))    
        if ML_data.shape == (self.N,self.J):
            ML_matrix_data = np.copy(ML_data)
        else:
            ML_matrix_data = ML_data.reshape((self.N,self.J))
        if ML_model.shape == (self.N,self.J):
            ML_matrix_model = np.copy(ML_model)
        else:
            ML_matrix_model = ML_model.reshape((self.N,self.J))
        if ML_error.shape == (self.N,self.J):
            ML_matrix_err = np.copy(ML_error)    
        else:
            ML_matrix_err = ML_error.reshape((self.N,self.J))  

        X = self.x.reshape((self.N,self.J))
        Y = self.y.reshape((self.N,self.J))

        fig = plt.figure()
        fig.set_figheight(20)
        fig.set_figwidth(25)

        if lim=='data':
            vmin = [np.min(brightness_data),np.min(v_data),np.min(sigma_data),np.min(ML_data)]
            vmax = [np.max(brightness_data),np.max(v_data),np.max(sigma_data),np.min(ML_data)]
        elif lim=='model':
            vmin = [np.min(brightness_model),np.min(v_model),np.min(sigma_model),np.min(ML_model)]
            vmax = [np.max(brightness_model),np.max(v_model),np.max(sigma_model),np.min(ML_model)]
        elif lim=='user':
            if vmin==None or vmax==None:
                print('error: must specify limits')
        else:
            print('lim can be data or model or user')
        if plot_ML == True: 
            ax1  = fig.add_subplot(3,4,1) 
            ax2  = fig.add_subplot(3,4,2) 
            ax3  = fig.add_subplot(3,4,3) 
            ax4  = fig.add_subplot(3,4,4) 
            ax5  = fig.add_subplot(3,4,5) 
            ax6  = fig.add_subplot(3,4,6) 
            ax7  = fig.add_subplot(3,4,7) 
            ax8  = fig.add_subplot(3,4,8) 
            ax9  = fig.add_subplot(3,4,9)
            ax10 = fig.add_subplot(3,4,10) 
            ax11 = fig.add_subplot(3,4,11) 
            ax12 = fig.add_subplot(3,4,12) 
            ax1.set_xlabel('x [arcsec]'),ax1.set_ylabel('y [arcsec]')
            ax2.set_xlabel('x [arcsec]'),ax2.set_ylabel('y [arcsec]')
            ax3.set_xlabel('x [arcsec]'),ax3.set_ylabel('y [arcsec]')
            ax4.set_xlabel('x [arcsec]'),ax4.set_ylabel('y [arcsec]')
            ax5.set_xlabel('x [arcsec]'),ax5.set_ylabel('y [arcsec]')
            ax6.set_xlabel('x [arcsec]'),ax6.set_ylabel('y [arcsec]')
            ax7.set_xlabel('x [arcsec]'),ax7.set_ylabel('y [arcsec]')
            ax8.set_xlabel('x [arcsec]'),ax8.set_ylabel('y [arcsec]')
            ax9.set_xlabel('x [arcsec]'),ax9.set_ylabel('y [arcsec]')
            ax10.set_xlabel('x [arcsec]'),ax10.set_ylabel('y [arcsec]')
            ax11.set_xlabel('x [arcsec]'),ax11.set_ylabel('y [arcsec]')
            ax12.set_xlabel('x [arcsec]'),ax12.set_ylabel('y [arcsec]')
            ax1.set_xlim(self.xmin,self.xmax),ax1.set_ylim(self.ymin,self.ymax)
            ax2.set_xlim(self.xmin,self.xmax),ax2.set_ylim(self.ymin,self.ymax)
            ax3.set_xlim(self.xmin,self.xmax),ax3.set_ylim(self.ymin,self.ymax)
            ax4.set_xlim(self.xmin,self.xmax),ax4.set_ylim(self.ymin,self.ymax)
            ax5.set_xlim(self.xmin,self.xmax),ax5.set_ylim(self.ymin,self.ymax)
            ax6.set_xlim(self.xmin,self.xmax),ax6.set_ylim(self.ymin,self.ymax)
            ax7.set_xlim(self.xmin,self.xmax),ax7.set_ylim(self.ymin,self.ymax)
            ax8.set_xlim(self.xmin,self.xmax),ax8.set_ylim(self.ymin,self.ymax)
            ax9.set_xlim(self.xmin,self.xmax),ax9.set_ylim(self.ymin,self.ymax)
            ax10.set_xlim(self.xmin,self.xmax),ax10.set_ylim(self.ymin,self.ymax)
            ax11.set_xlim(self.xmin,self.xmax),ax11.set_ylim(self.ymin,self.ymax)
            ax12.set_xlim(self.xmin,self.xmax),ax12.set_ylim(self.ymin,self.ymax)
            ax1.set_aspect('equal', 'box')  
            ax2.set_aspect('equal', 'box')  
            ax3.set_aspect('equal', 'box')  
            ax4.set_aspect('equal', 'box')  
            ax5.set_aspect('equal', 'box')  
            ax6.set_aspect('equal', 'box')  
            ax7.set_aspect('equal', 'box')  
            ax8.set_aspect('equal', 'box')  
            ax9.set_aspect('equal', 'box')  
            ax10.set_aspect('equal', 'box')  
            ax11.set_aspect('equal', 'box')  
            ax12.set_aspect('equal', 'box')  

            # BRIGHTNESS
            pmesh1 = ax1.pcolormesh(X,Y,B_matrix_data,cmap = 'viridis',shading='nearest',vmin=vmin[0],vmax=vmax[0])
            cbar1  = plt.colorbar(pmesh1,ax=ax1,label=r'$log_{10}(B)$ $[L_{\odot}/arcsec^2]$')
            #B_lim  = pmesh1.get_clim()
            pmesh5 = ax5.pcolormesh(X,Y,B_matrix_model,cmap = 'viridis',shading='nearest',vmin=vmin[0],vmax=vmax[0])
            cbar5  = plt.colorbar(pmesh5,ax=ax5,label=r'$log_{10}(B)$ $[L_{\odot}/arcsec^2]$' )
            #pmesh5.set_clim(B_lim)
            pmesh9 = ax9.pcolormesh(X,Y,(10**B_matrix_data-10**B_matrix_model)/10**B_matrix_data,cmap = 'viridis',shading='nearest')
            cbar9  = plt.colorbar(pmesh9,ax=ax9,label=r'$(B_{data}-B_{model})$ / $B_{data}$')
            pmesh9.set_clim((-1.,1.))

            # VELOCITY
            pmesh2 = ax2.pcolormesh(X,Y,v_matrix_data,cmap = 'viridis',shading='nearest',vmin=vmin[1],vmax=vmax[1])
            cbar2  = plt.colorbar(pmesh2,ax=ax2,label=r'$v_{los}$ $[km/s]$')
            #v_lim  = pmesh2.get_clim()
            pmesh6 = ax6.pcolormesh(X,Y,v_matrix_model,cmap = 'viridis',shading='nearest',vmin=vmin[1],vmax=vmax[1])
            cbar6  = plt.colorbar(pmesh6,ax=ax6,label=r'$v_{los}$ $[km/s]$')
            #pmesh6.set_clim((v_lim))
            pmesh10 = ax10.pcolormesh(X,Y,(v_matrix_data-v_matrix_model)/v_matrix_err,cmap = 'viridis',shading='nearest')
            cbar10  = plt.colorbar(pmesh10,ax=ax10,label=r'$(v_{data}-v_{model})$ / $\delta v$')
            pmesh10.set_clim((-10.,10.))

            # SIGMA
            pmesh3 = ax3.pcolormesh(X,Y,sigma_matrix_data,cmap = 'viridis',shading='nearest',vmin=vmin[2],vmax=vmax[2])
            cbar3  = plt.colorbar(pmesh3,ax=ax3,label=r'$\sigma_{los}$ $[km/s]$')
            #sigma_lim  = pmesh3.get_clim()
            pmesh7 = ax7.pcolormesh(X,Y,sigma_matrix_model,cmap = 'viridis',shading='nearest',vmin=vmin[2],vmax=vmax[2])
            cbar7  = plt.colorbar(pmesh7,ax=ax7,label=r'$\sigma_{los}$ $[km/s]$')
            #pmesh7.set_clim((sigma_lim))
            pmesh11 = ax11.pcolormesh(X,Y,(sigma_matrix_data-sigma_matrix_model)/sigma_matrix_err,cmap = 'viridis',shading='nearest')
            cbar11  = plt.colorbar(pmesh11,ax=ax11,label=r'$(\sigma_{data}-\sigma_{model})$ / $\delta \sigma$')
            pmesh11.set_clim((-10.,10.))

            # ML ratio
            pmesh4 = ax4.pcolormesh(X,Y,ML_matrix_data,cmap = 'viridis',shading='nearest',vmin=vmin[3],vmax=vmax[3])
            cbar4  = plt.colorbar(pmesh4,ax=ax4,label=r'$ML$ $[M_{\odot}/L_{\odot}]$')
            #ML_lim  = pmesh4.get_clim()
            pmesh8 = ax8.pcolormesh(X,Y,ML_matrix_model,cmap = 'viridis',shading='nearest',vmin=vmin[3],vmax=vmax[3])
            cbar8  = plt.colorbar(pmesh8,ax=ax8,label=r'$ML$ $[M_{\odot}/L_{\odot}]$')
            #pmesh8.set_clim((ML_lim))
            pmesh12 = ax12.pcolormesh(X,Y,(ML_matrix_data-ML_matrix_model)/ML_matrix_err,cmap = 'viridis',shading='nearest')
            cbar12  = plt.colorbar(pmesh12,ax=ax12,label=r'$(ML_{data}-ML_{model})$ / $\delta ML$')
            pmesh12.set_clim((-10.,10.))

            plt.subplots_adjust(top=0.99,
                                bottom=0.07,
                                left=0.005,
                                right=0.98,
                                hspace=0.305,
                                wspace=0.18)

        else: 
            ax1  = fig.add_subplot(3,3,1)
            ax2  = fig.add_subplot(3,3,2)
            ax3  = fig.add_subplot(3,3,3)
            ax4  = fig.add_subplot(3,3,4)
            ax5  = fig.add_subplot(3,3,5)
            ax6  = fig.add_subplot(3,3,6)
            ax7  = fig.add_subplot(3,3,7)
            ax8  = fig.add_subplot(3,3,8)
            ax9  = fig.add_subplot(3,3,9)
        
            ax1.set_xlabel('x [arcsec]'),ax1.set_ylabel('y [arcsec]')
            ax2.set_xlabel('x [arcsec]'),ax2.set_ylabel('y [arcsec]')
            ax3.set_xlabel('x [arcsec]'),ax3.set_ylabel('y [arcsec]')
            ax4.set_xlabel('x [arcsec]'),ax4.set_ylabel('y [arcsec]')
            ax5.set_xlabel('x [arcsec]'),ax5.set_ylabel('y [arcsec]')
            ax6.set_xlabel('x [arcsec]'),ax6.set_ylabel('y [arcsec]')
            ax7.set_xlabel('x [arcsec]'),ax7.set_ylabel('y [arcsec]')
            ax8.set_xlabel('x [arcsec]'),ax8.set_ylabel('y [arcsec]')
            ax9.set_xlabel('x [arcsec]'),ax9.set_ylabel('y [arcsec]')
            ax1.set_xlim(self.xmin,self.xmax),ax1.set_ylim(self.ymin,self.ymax)
            ax2.set_xlim(self.xmin,self.xmax),ax2.set_ylim(self.ymin,self.ymax)
            ax3.set_xlim(self.xmin,self.xmax),ax3.set_ylim(self.ymin,self.ymax)
            ax4.set_xlim(self.xmin,self.xmax),ax4.set_ylim(self.ymin,self.ymax)
            ax5.set_xlim(self.xmin,self.xmax),ax5.set_ylim(self.ymin,self.ymax)
            ax6.set_xlim(self.xmin,self.xmax),ax6.set_ylim(self.ymin,self.ymax)
            ax7.set_xlim(self.xmin,self.xmax),ax7.set_ylim(self.ymin,self.ymax)
            ax8.set_xlim(self.xmin,self.xmax),ax8.set_ylim(self.ymin,self.ymax)
            ax9.set_xlim(self.xmin,self.xmax),ax9.set_ylim(self.ymin,self.ymax)
            ax1.set_aspect('equal', 'box')  
            ax2.set_aspect('equal', 'box')  
            ax3.set_aspect('equal', 'box')  
            ax4.set_aspect('equal', 'box')  
            ax5.set_aspect('equal', 'box')  
            ax6.set_aspect('equal', 'box')  
            ax7.set_aspect('equal', 'box')  
            ax8.set_aspect('equal', 'box')  
            ax9.set_aspect('equal', 'box') 
            # BRIGHTNESS
            pmesh1 = ax1.pcolormesh(X,Y,B_matrix_data,cmap = 'viridis',shading='nearest',vmin=vmin[0],vmax=vmax[0])
            cbar1  = plt.colorbar(pmesh1,ax=ax1,label=r'$log_{10}(B)$ $[L_{\odot}/arcsec^2]$')
            #B_lim  = pmesh1.get_clim()
            pmesh4 = ax4.pcolormesh(X,Y,B_matrix_model,cmap = 'viridis',shading='nearest',vmin=vmin[0],vmax=vmax[0])
            cbar4  = plt.colorbar(pmesh4,ax=ax4,label=r'$log_{10}(B)$ $[L_{\odot}/arcsec^2]$')
            #pmesh4.set_clim(B_lim)
            pmesh7 = ax7.pcolormesh(X,Y,(10**B_matrix_data-10**B_matrix_model)/10**B_matrix_data,cmap = 'viridis',shading='nearest')
            cbar7  = plt.colorbar(pmesh7,ax=ax7,label=r'$(B_{data}-B_{model})$ / $B_{data}$')
            pmesh7.set_clim((-1.,1.))

            # VELOCITY
            pmesh2 = ax2.pcolormesh(X,Y,v_matrix_data,cmap = 'viridis',shading='nearest',vmin=vmin[1],vmax=vmax[1])
            cbar2  = plt.colorbar(pmesh2,ax=ax2,label=r'$v_{los}$ $[km/s]$')
            #v_lim  = pmesh2.get_clim()
            pmesh5 = ax5.pcolormesh(X,Y,v_matrix_model,cmap = 'viridis',shading='nearest',vmin=vmin[1],vmax=vmax[1])
            cbar5  = plt.colorbar(pmesh5,ax=ax5,label=r'$v_{los}$ $[km/s]$')
            #pmesh5.set_clim((v_lim))
            pmesh8 = ax8.pcolormesh(X,Y,(v_matrix_data-v_matrix_model)/v_matrix_err,cmap = 'viridis',shading='nearest')
            cbar8  = plt.colorbar(pmesh8,ax=ax8,label=r'$(v_{data}-v_{model})$ / $\delta v$')
            pmesh8.set_clim((-10.,10.))

            # SIGMA
            pmesh3 = ax3.pcolormesh(X,Y,sigma_matrix_data,cmap = 'viridis',shading='nearest',vmin=vmin[2],vmax=vmax[2])
            cbar3  = plt.colorbar(pmesh3,ax=ax3,label=r'$\sigma_{los}$ $[km/s]$')
            #sigma_lim  = pmesh3.get_clim()
            pmesh6 = ax6.pcolormesh(X,Y,sigma_matrix_model,cmap = 'viridis',shading='nearest',vmin=vmin[2],vmax=vmax[2])
            cbar6  = plt.colorbar(pmesh6,ax=ax6,label=r'$\sigma_{los}$ $[km/s]$')
            #pmesh6.set_clim((sigma_lim))
            pmesh9 = ax9.pcolormesh(X,Y,(sigma_matrix_data-sigma_matrix_model)/sigma_matrix_err,cmap = 'viridis',shading='nearest')
            cbar9  = plt.colorbar(pmesh9,ax=ax9,label=r'$(\sigma_{data}-\sigma_{model})$ / $\delta \sigma$')
            pmesh9.set_clim((-10.,10.))
        
            plt.subplots_adjust(top=0.98,
                                bottom=0.075,
                                left=0.1,
                                right=0.905,
                                hspace=0.29,
                                wspace=0.2)

        return fig


    def all_1D_profile(self,
                     arcsec_to_kpc,
                     r_avg_brightness,
                     brightness_avg_data,
                     brightness_avg_model,
                     r_avg_v,
                     v_avg_data,
                     v_avg_model,
                     r_avg_sigma,
                     sigma_avg_data,
                     sigma_avg_model,
                     r_avg_ML,
                     ML_avg_data,
                     ML_avg_model,
                     brightness_avg_error,
                     v_avg_error,
                     sigma_avg_error,
                     ML_avg_error,
                     plot_ML=True):
        fig = plt.figure()
        fig.set_figheight(20)
        fig.set_figwidth(25)
        if plot_ML==True:
            ax1  = fig.add_subplot(1,4,1)
            ax2  = fig.add_subplot(1,4,2)
            ax3  = fig.add_subplot(1,4,3)
            ax4  = fig.add_subplot(1,4,4)
            #ax1.set_aspect('equal', 'box')
            #ax2.set_aspect('equal', 'box')
            #ax3.set_aspect('equal', 'box')
            #ax4.set_aspect('equal', 'box')
            
            ax1.plot((r_avg_brightness),(brightness_avg_data),'ok')
            ax1.fill_between((r_avg_brightness), ((brightness_avg_data-brightness_avg_error)), ((brightness_avg_data+brightness_avg_error)),color='black',alpha=0.5)
            ax1.plot((r_avg_brightness),(brightness_avg_model),'b',lw=3)
            ax1.set_xscale('log')
            ax1.set_yscale('log')
            ax1.set_xlabel('r [arcsec]')#,size=10.5)
            ax1.set_ylabel(r'$B$ $[L_{\odot}/kpc^2]$')#,size=10.5)
            
            ax2.plot((r_avg_v),(v_avg_data),'ok',label='data')
            ax2.fill_between((r_avg_v), ((v_avg_data-v_avg_error)), ((v_avg_data+v_avg_error)),color='black',alpha=0.5)
            ax2.plot((r_avg_v),(v_avg_model),'b',label='Fabio',lw=3)
            ax2.set_xlabel('r [arcsec]')#,size=10.5)
            ax2.set_ylabel(r'$v$ [km/s]')#,size=10.5)
            
            ax3.plot((r_avg_sigma),(sigma_avg_data),'ok',label='data')
            ax3.fill_between((r_avg_sigma), ((sigma_avg_data-sigma_avg_error)), ((sigma_avg_data+sigma_avg_error)),color='black',alpha=0.5)
            ax3.plot((r_avg_sigma),(sigma_avg_model),'b',label='Fabio',lw=3)
            ax3.set_xlabel('r [arcsec]')#,size=10.5)
            ax3.set_ylabel(r'$\sigma$ [km/s]')#,size=10.5)
            
            ax4.plot((r_avg_ML),(ML_avg_data),'ok',label='data')
            ax4.fill_between((r_avg_ML), ((ML_avg_data-ML_avg_error)), ((ML_avg_data+ML_avg_error)),color='black',alpha=0.5)
            ax4.plot((r_avg_ML),(ML_avg_model),'b',label='Fabio',lw=3)
            ax4.set_xlabel('r [arcsec]')#,size=10.5)
            ax4.set_ylabel(r'$M/L$ $[M_{\odot}/L_{\odot}]$')#,size=10.5)
            
            plt.subplots_adjust(top=0.74,
                                bottom=0.305,
                                left=0.06,
                                right=0.99,
                                hspace=0.2,
                                wspace=0.33)
        
        else:
            ax1  = fig.add_subplot(1,3,1)
            ax2  = fig.add_subplot(1,3,2)
            ax3  = fig.add_subplot(1,3,3)
            #ax1.set_aspect('equal', 'box')
            #ax2.set_aspect('equal', 'box')
            #ax3.set_aspect('equal', 'box')
            
            ax1.plot((r_avg_brightness),(brightness_avg_data),'ok')
            ax1.fill_between((r_avg_brightness), ((brightness_avg_data-brightness_avg_error)), ((brightness_avg_data+brightness_avg_error)),color='black',alpha=0.5)
            ax1.plot((r_avg_brightness),(brightness_avg_model),'b',lw=3)
            ax1.set_xscale('log')
            ax1.set_yscale('log')
            ax1.set_xlabel('r [arcsec]')#,size=10.5)
            ax1.set_ylabel(r'$B$ $[L_{\odot}/kpc^2]$')#,size=10.5)
            ax2.plot((r_avg_v),(v_avg_data),'ok',label='data')
            ax2.fill_between((r_avg_v), ((v_avg_data-v_avg_error)), ((v_avg_data+v_avg_error)),color='black',alpha=0.5)
            ax2.plot((r_avg_v),(v_avg_model),'b',label='Fabio',lw=3)
            ax2.set_xlabel('r [arcsec]')#,size=10.5)
            ax2.set_ylabel(r'$v$ [km/s]')#,size=10.5)
            ax3.plot((r_avg_sigma),(sigma_avg_data),'ok',label='data')
            ax3.fill_between((r_avg_sigma), ((sigma_avg_data-sigma_avg_error)), ((sigma_avg_data+sigma_avg_error)),color='black',alpha=0.5)
            ax3.plot((r_avg_sigma),(sigma_avg_model),'b',label='Fabio',lw=3)
            ax3.set_xlabel('r [arcsec]')#,size=10.5)
            ax3.set_ylabel(r'$\sigma$ [km/s]')#,size=10.5)

            plt.subplots_adjust(top=0.79,
                                bottom=0.2,
                                left=0.06,
                                right=0.985,
                                hspace=0.245,
                                wspace=0.3)

        return fig


    def all_1D_profile_all_outputs(self,
                     arcsec_to_kpc,
                     r_avg_brightness,
                     brightness_avg_data,
                     brightness_avg_model,
                     B_B_avg,
                     B_D1_avg,
                     B_D2_avg,
                     r_avg_v,
                     v_avg_data,
                     v_avg_model,
                     v_B_avg,
                     v_D1_avg,
                     v_D2_avg,
                     r_avg_sigma,
                     sigma_avg_data,
                     sigma_avg_model,
                     sigma_B_avg,
                     sigma_D1_avg,
                     sigma_D2_avg,
                     r_avg_ML,
                     ML_avg_data,
                     ML_avg_model,
                     ML_B_avg,
                     ML_D1_avg,
                     ML_D2_avg,
                     brightness_avg_error,
                     v_avg_error,
                     sigma_avg_error,
                     ML_avg_error,
                     plot_ML=True):
        fig = plt.figure()
        fig.set_figheight(20)
        fig.set_figwidth(25)
        if plot_ML==True:
            ax1  = fig.add_subplot(1,4,1)
            ax2  = fig.add_subplot(1,4,2)
            ax3  = fig.add_subplot(1,4,3)
            ax4  = fig.add_subplot(1,4,4)
            #ax1.set_aspect('equal', 'box')
            #ax2.set_aspect('equal', 'box')
            #ax3.set_aspect('equal', 'box')
            #ax4.set_aspect('equal', 'box')
            
            ax1.plot(np.log10(r_avg_brightness),(brightness_avg_data),'ok')
            ax1.fill_between(np.log10(r_avg_brightness), ((brightness_avg_data-brightness_avg_error)), ((brightness_avg_data+brightness_avg_error)),color='black',alpha=0.5)
            ax1.plot(np.log10(r_avg_brightness),(brightness_avg_model),'b',lw=3)
            ax1.plot(np.log10(r_avg_brightness),(B_B_avg),'m' ,lw=3)
            ax1.plot(np.log10(r_avg_brightness),(B_D1_avg),'r',lw=3)
            ax1.plot(np.log10(r_avg_brightness),(B_D2_avg),'g',lw=3)
            #ax1.set_xscale('log')
            #ax1.set_yscale('log')
            ax1.set_xlabel(r'$log_{10}{r}$ [arcsec]')#,size=10.5)
            ax1.set_ylabel(r'$log_{10}{B}$ $[L_{\odot}/kpc^2]$')#,size=10.5)
            ax1_sec = ax1.secondary_xaxis('top')
            ax1_sec.set_xticks(np.log10(10**r_avg_brightness*arcsec_to_kpc))
            ax1_sec.set_xlabel(r'$log_{10}{r}$ [kpc]')

            ax2.plot((r_avg_v),(v_avg_data),'ok',label='data')
            ax2.fill_between((r_avg_v), ((v_avg_data-v_avg_error)), ((v_avg_data+v_avg_error)),color='black',alpha=0.5)
            ax2.plot((r_avg_v),(v_avg_model),'b',lw=3)
            ax2.plot((r_avg_v),(v_B_avg),'m',lw=3)
            ax2.plot((r_avg_v),(v_D1_avg),'r',lw=3)
            ax2.plot((r_avg_v),(v_D2_avg),'g',lw=3)
            ax2.set_xlabel('r [arcsec]')#,size=10.5)
            ax2.set_ylabel(r'$v$ [km/s]')#,size=10.5)
            ax2_sec = ax3.secondary_xaxis('top')
            ax2_sec.set_xticks(r_avg_v*arcsec_to_kpc)
            ax2_sec.set_xlabel('r [kpc]')

            ax3.plot((r_avg_sigma),(sigma_avg_data),'ok',label='data')
            ax3.fill_between((r_avg_sigma), ((sigma_avg_data-sigma_avg_error)), ((sigma_avg_data+sigma_avg_error)),color='black',alpha=0.5)
            ax3.plot((r_avg_sigma),(sigma_avg_model),'b',lw=3)
            ax3.plot((r_avg_sigma),(sigma_B_avg),'m',lw=3)
            ax3.plot((r_avg_sigma),(sigma_D1_avg),'r',lw=3)
            ax3.plot((r_avg_sigma),(sigma_D2_avg),'g',lw=3)
            ax3.set_xlabel('r [arcsec]')#,size=10.5)
            ax3.set_ylabel(r'$\sigma$ [km/s]')#,size=10.5)
            ax3_sec = ax3.secondary_xaxis('top')
            ax3_sec.set_xticks(r_avg_sigma*arcsec_to_kpc)
            ax3_sec.set_xlabel('r [kpc]')

            ax4.plot((r_avg_ML),(ML_avg_data),'ok',label='data')
            ax4.fill_between((r_avg_ML), ((ML_avg_data-ML_avg_error)), ((ML_avg_data+ML_avg_error)),color='black',alpha=0.5)
            ax4.plot((r_avg_ML),(ML_avg_model),'b',lw=3)
            ax4.plot((r_avg_ML),(ML_B_avg),'m',    lw=3,label='B')
            ax4.plot((r_avg_ML),(ML_D1_avg),'r',   lw=3,label='D1')
            ax4.plot((r_avg_ML),(ML_D2_avg),'g',   lw=3,label='D2')
            ax4.set_xlabel('r [arcsec]')#,size=10.5)
            ax4.set_ylabel(r'$M/L$ $[M_{\odot}/L_{\odot}]$')#,size=10.5)
            ax4_sec = ax4.secondary_xaxis('top')
            ax4_sec.set_xticks(arcsec_to_kpc*r_avg_ML)
            ax4_sec.set_xlabel('r [kpc]')

            ax4.legend(loc='best')
            plt.subplots_adjust(top=0.74,
                                bottom=0.305,
                                left=0.06,
                                right=0.99,
                                hspace=0.2,
                                wspace=0.33)
        
        else:
            ax1  = fig.add_subplot(1,3,1)
            ax2  = fig.add_subplot(1,3,2)
            ax3  = fig.add_subplot(1,3,3)
            #ax1.set_aspect('equal', 'box')
            #ax2.set_aspect('equal', 'box')
            #ax3.set_aspect('equal', 'box')
            
            ax1.plot((r_avg_brightness),(brightness_avg_data),'ok')
            ax1.fill_between((r_avg_brightness), ((brightness_avg_data-brightness_avg_error)), ((brightness_avg_data+brightness_avg_error)),color='black',alpha=0.5)
            ax1.plot((r_avg_brightness),(brightness_avg_model),'b',lw=3)
            ax1.set_xscale('log')
            ax1.set_yscale('log')
            ax1.set_xlabel('r [arcsec]')#,size=10.5)
            ax1.set_ylabel(r'$B$ $[L_{\odot}/kpc^2]$')#,size=10.5)
            ax2.plot((r_avg_v),(v_avg_data),'ok',label='data')
            ax2.fill_between((r_avg_v), ((v_avg_data-v_avg_error)), ((v_avg_data+v_avg_error)),color='black',alpha=0.5)
            ax2.plot((r_avg_v),(v_avg_model),'b',label='Fabio',lw=3)
            ax2.set_xlabel('r [arcsec]')#,size=10.5)
            ax2.set_ylabel(r'$v$ [km/s]')#,size=10.5)
            ax3.plot((r_avg_sigma),(sigma_avg_data),'ok',label='data')
            ax3.fill_between((r_avg_sigma), ((sigma_avg_data-sigma_avg_error)), ((sigma_avg_data+sigma_avg_error)),color='black',alpha=0.5)
            ax3.plot((r_avg_sigma),(sigma_avg_model),'b',label='Fabio',lw=3)
            ax3.set_xlabel('r [arcsec]')#,size=10.5)
            ax3.set_ylabel(r'$\sigma$ [km/s]')#,size=10.5)

            plt.subplots_adjust(top=0.79,
                                bottom=0.2,
                                left=0.06,
                                right=0.985,
                                hspace=0.245,
                                wspace=0.3)

        return fig

if __name__ == '__main__':

    x,y = np.linspace(-10,10,101),np.linspace(-10,10,101) 
    x,y = np.meshgrid(x,y)#,indexing='ij')
    x,y = x.ravel(),y.ravel()

    z = np.random.normal(1,0.1,x.shape)

    d1 = data(x,y)
    x_g,y_g,psf,K,indexing = d1.refined_grid_indexing(sigma_psf=1.321)
    x_g1,y_g1,psf1,K1,indexing1 = d1.refined_grid_var_resolution(sigma_psf=1.321,N_px=10,resolution=2)
    
    #x_g1,y_g1,psf1,K1 = d1.refined_grid(sigma_psf=1.321)



    fig = d1.all_maps_kpc(z,
                          z,
                          z,
                          z,
                          z,
                          z,
                          z,
                          z,
                          z,
                          z,
                          z,
                          False)
    fig.show()

    plt.show()